static const AVInputFormat * const demuxer_list[] = {
    &ff_aac_demuxer,
    &ff_flac_demuxer,
    &ff_matroska_demuxer,
    &ff_mov_demuxer,
    &ff_mp3_demuxer,
    &ff_ogg_demuxer,
    &ff_wav_demuxer,
    NULL };
