static const AVCodecParser * const parser_list[] = {
    &ff_aac_parser,
    &ff_flac_parser,
    &ff_h264_parser,
    &ff_mpegaudio_parser,
    &ff_opus_parser,
    &ff_vorbis_parser,
    &ff_vp3_parser,
    &ff_vp8_parser,
    NULL };
