/**
 * Copyright (C) 2026 Lynne
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef SWSCALE_VULKAN_OPS_H
#define SWSCALE_VULKAN_OPS_H

#include "libavutil/vulkan.h"
#include "../swscale.h"

#if CONFIG_LIBSHADERC || CONFIG_LIBGLSLANG
#include "libavutil/vulkan_spirv.h"
#endif

typedef struct FFVulkanOpsCtx {
    FFVulkanContext vkctx;
    AVVulkanDeviceQueueFamily *qf;
    FFVkExecPool e;
#if CONFIG_LIBSHADERC || CONFIG_LIBGLSLANG
    FFVkSPIRVCompiler *spvc;
#endif
} FFVulkanOpsCtx;

int ff_sws_vk_init(SwsContext *sws, AVBufferRef *dev_ref);

#endif /* SWSCALE_VULKAN_OPS_H */
