/* Automatically generated by configure - do not modify! */
#ifndef FFMPEG_CONFIG_COMPONENTS_H
#define FFMPEG_CONFIG_COMPONENTS_H
#define CONFIG_AAC_ADTSTOASC_BSF 0
#define CONFIG_AHX_TO_MP2_BSF 0
#define CONFIG_APV_METADATA_BSF 0
#define CONFIG_AV1_FRAME_MERGE_BSF 0
#define CONFIG_AV1_FRAME_SPLIT_BSF 0
#define CONFIG_AV1_METADATA_BSF 0
#define CONFIG_CHOMP_BSF 0
#define CONFIG_DUMP_EXTRADATA_BSF 0
#define CONFIG_DCA_CORE_BSF 0
#define CONFIG_DOVI_RPU_BSF 0
#define CONFIG_DTS2PTS_BSF 0
#define CONFIG_DV_ERROR_MARKER_BSF 0
#define CONFIG_EAC3_CORE_BSF 0
#define CONFIG_EIA608_TO_SMPTE436M_BSF 0
#define CONFIG_EVC_FRAME_MERGE_BSF 0
#define CONFIG_EXTRACT_EXTRADATA_BSF 0
#define CONFIG_FILTER_UNITS_BSF 0
#define CONFIG_H264_METADATA_BSF 0
#define CONFIG_H264_MP4TOANNEXB_BSF 0
#define CONFIG_H264_REDUNDANT_PPS_BSF 0
#define CONFIG_HAPQA_EXTRACT_BSF 0
#define CONFIG_HEVC_METADATA_BSF 0
#define CONFIG_HEVC_MP4TOANNEXB_BSF 0
#define CONFIG_IMX_DUMP_HEADER_BSF 0
#define CONFIG_LCEVC_METADATA_BSF 0
#define CONFIG_MEDIA100_TO_MJPEGB_BSF 0
#define CONFIG_MJPEG2JPEG_BSF 0
#define CONFIG_MJPEGA_DUMP_HEADER_BSF 0
#define CONFIG_MPEG2_METADATA_BSF 0
#define CONFIG_MPEG4_UNPACK_BFRAMES_BSF 0
#define CONFIG_MOV2TEXTSUB_BSF 0
#define CONFIG_NOISE_BSF 0
#define CONFIG_NULL_BSF 0
#define CONFIG_OPUS_METADATA_BSF 0
#define CONFIG_PCM_RECHUNK_BSF 0
#define CONFIG_PGS_FRAME_MERGE_BSF 0
#define CONFIG_PRORES_METADATA_BSF 0
#define CONFIG_REMOVE_EXTRADATA_BSF 0
#define CONFIG_SETTS_BSF 0
#define CONFIG_SHOWINFO_BSF 0
#define CONFIG_SMPTE436M_TO_EIA608_BSF 0
#define CONFIG_TEXT2MOVSUB_BSF 0
#define CONFIG_TRACE_HEADERS_BSF 0
#define CONFIG_TRUEHD_CORE_BSF 0
#define CONFIG_VP9_METADATA_BSF 0
#define CONFIG_VP9_RAW_REORDER_BSF 0
#define CONFIG_VP9_SUPERFRAME_BSF 0
#define CONFIG_VP9_SUPERFRAME_SPLIT_BSF 0
#define CONFIG_VVC_METADATA_BSF 0
#define CONFIG_VVC_MP4TOANNEXB_BSF 0
#define CONFIG_AASC_DECODER 0
#define CONFIG_AIC_DECODER 0
#define CONFIG_ALIAS_PIX_DECODER 0
#define CONFIG_AGM_DECODER 0
#define CONFIG_AMV_DECODER 0
#define CONFIG_ANM_DECODER 0
#define CONFIG_ANSI_DECODER 0
#define CONFIG_APNG_DECODER 0
#define CONFIG_APV_DECODER 0
#define CONFIG_ARBC_DECODER 0
#define CONFIG_ARGO_DECODER 0
#define CONFIG_ASV1_DECODER 0
#define CONFIG_ASV2_DECODER 0
#define CONFIG_AURA_DECODER 0
#define CONFIG_AURA2_DECODER 0
#define CONFIG_AVRP_DECODER 0
#define CONFIG_AVRN_DECODER 0
#define CONFIG_AVS_DECODER 0
#define CONFIG_AVUI_DECODER 0
#define CONFIG_BETHSOFTVID_DECODER 0
#define CONFIG_BFI_DECODER 0
#define CONFIG_BINK_DECODER 0
#define CONFIG_BITPACKED_DECODER 0
#define CONFIG_BMP_DECODER 0
#define CONFIG_BMV_VIDEO_DECODER 0
#define CONFIG_BRENDER_PIX_DECODER 0
#define CONFIG_C93_DECODER 0
#define CONFIG_CAVS_DECODER 0
#define CONFIG_CDGRAPHICS_DECODER 0
#define CONFIG_CDTOONS_DECODER 0
#define CONFIG_CDXL_DECODER 0
#define CONFIG_CFHD_DECODER 0
#define CONFIG_CINEPAK_DECODER 0
#define CONFIG_CLEARVIDEO_DECODER 0
#define CONFIG_CLJR_DECODER 0
#define CONFIG_CLLC_DECODER 0
#define CONFIG_COMFORTNOISE_DECODER 0
#define CONFIG_CPIA_DECODER 0
#define CONFIG_CRI_DECODER 0
#define CONFIG_CSCD_DECODER 0
#define CONFIG_CYUV_DECODER 0
#define CONFIG_DDS_DECODER 0
#define CONFIG_DFA_DECODER 0
#define CONFIG_DIRAC_DECODER 0
#define CONFIG_DNXHD_DECODER 0
#define CONFIG_DPX_DECODER 0
#define CONFIG_DSICINVIDEO_DECODER 0
#define CONFIG_DVAUDIO_DECODER 0
#define CONFIG_DVVIDEO_DECODER 0
#define CONFIG_DXA_DECODER 0
#define CONFIG_DXTORY_DECODER 0
#define CONFIG_DXV_DECODER 0
#define CONFIG_EACMV_DECODER 0
#define CONFIG_EAMAD_DECODER 0
#define CONFIG_EATGQ_DECODER 0
#define CONFIG_EATGV_DECODER 0
#define CONFIG_EATQI_DECODER 0
#define CONFIG_EIGHTBPS_DECODER 0
#define CONFIG_EIGHTSVX_EXP_DECODER 0
#define CONFIG_EIGHTSVX_FIB_DECODER 0
#define CONFIG_ESCAPE124_DECODER 0
#define CONFIG_ESCAPE130_DECODER 0
#define CONFIG_EXR_DECODER 0
#define CONFIG_FFV1_DECODER 0
#define CONFIG_FFVHUFF_DECODER 0
#define CONFIG_FIC_DECODER 0
#define CONFIG_FITS_DECODER 0
#define CONFIG_FLASHSV_DECODER 0
#define CONFIG_FLASHSV2_DECODER 0
#define CONFIG_FLIC_DECODER 0
#define CONFIG_FLV_DECODER 0
#define CONFIG_FMVC_DECODER 0
#define CONFIG_FOURXM_DECODER 0
#define CONFIG_FRAPS_DECODER 0
#define CONFIG_FRWU_DECODER 0
#define CONFIG_G2M_DECODER 0
#define CONFIG_GDV_DECODER 0
#define CONFIG_GEM_DECODER 0
#define CONFIG_GIF_DECODER 0
#define CONFIG_H261_DECODER 0
#define CONFIG_H263_DECODER 0
#define CONFIG_H263I_DECODER 0
#define CONFIG_H263P_DECODER 0
#define CONFIG_H263_V4L2M2M_DECODER 0
#define CONFIG_H264_DECODER 0
#define CONFIG_H264_V4L2M2M_DECODER 0
#define CONFIG_H264_MEDIACODEC_DECODER 0
#define CONFIG_H264_MMAL_DECODER 0
#define CONFIG_H264_QSV_DECODER 0
#define CONFIG_H264_RKMPP_DECODER 0
#define CONFIG_HAP_DECODER 0
#define CONFIG_HEVC_DECODER 0
#define CONFIG_HEVC_QSV_DECODER 0
#define CONFIG_HEVC_RKMPP_DECODER 0
#define CONFIG_HEVC_V4L2M2M_DECODER 0
#define CONFIG_HNM4_VIDEO_DECODER 0
#define CONFIG_HQ_HQA_DECODER 0
#define CONFIG_HQX_DECODER 0
#define CONFIG_HUFFYUV_DECODER 0
#define CONFIG_HYMT_DECODER 0
#define CONFIG_IDCIN_DECODER 0
#define CONFIG_IFF_ILBM_DECODER 0
#define CONFIG_IMM4_DECODER 0
#define CONFIG_IMM5_DECODER 0
#define CONFIG_INDEO2_DECODER 0
#define CONFIG_INDEO3_DECODER 0
#define CONFIG_INDEO4_DECODER 0
#define CONFIG_INDEO5_DECODER 0
#define CONFIG_INTERPLAY_VIDEO_DECODER 0
#define CONFIG_IPU_DECODER 0
#define CONFIG_JPEG2000_DECODER 0
#define CONFIG_JPEGLS_DECODER 0
#define CONFIG_JV_DECODER 0
#define CONFIG_KGV1_DECODER 0
#define CONFIG_KMVC_DECODER 0
#define CONFIG_LAGARITH_DECODER 0
#define CONFIG_LEAD_DECODER 0
#define CONFIG_LOCO_DECODER 0
#define CONFIG_LSCR_DECODER 0
#define CONFIG_M101_DECODER 0
#define CONFIG_MAGICYUV_DECODER 0
#define CONFIG_MDEC_DECODER 0
#define CONFIG_MEDIA100_DECODER 0
#define CONFIG_MIMIC_DECODER 0
#define CONFIG_MJPEG_DECODER 0
#define CONFIG_MJPEGB_DECODER 0
#define CONFIG_MMVIDEO_DECODER 0
#define CONFIG_MOBICLIP_DECODER 0
#define CONFIG_MOTIONPIXELS_DECODER 0
#define CONFIG_MPEG1VIDEO_DECODER 0
#define CONFIG_MPEG2VIDEO_DECODER 0
#define CONFIG_MPEG4_DECODER 0
#define CONFIG_MPEG4_V4L2M2M_DECODER 0
#define CONFIG_MPEG4_MMAL_DECODER 0
#define CONFIG_MPEGVIDEO_DECODER 0
#define CONFIG_MPEG1_V4L2M2M_DECODER 0
#define CONFIG_MPEG2_MMAL_DECODER 0
#define CONFIG_MPEG2_V4L2M2M_DECODER 0
#define CONFIG_MPEG2_QSV_DECODER 0
#define CONFIG_MPEG2_MEDIACODEC_DECODER 0
#define CONFIG_MSA1_DECODER 0
#define CONFIG_MSCC_DECODER 0
#define CONFIG_MSMPEG4V1_DECODER 0
#define CONFIG_MSMPEG4V2_DECODER 0
#define CONFIG_MSMPEG4V3_DECODER 0
#define CONFIG_MSP2_DECODER 0
#define CONFIG_MSRLE_DECODER 0
#define CONFIG_MSS1_DECODER 0
#define CONFIG_MSS2_DECODER 0
#define CONFIG_MSVIDEO1_DECODER 0
#define CONFIG_MSZH_DECODER 0
#define CONFIG_MTS2_DECODER 0
#define CONFIG_MV30_DECODER 0
#define CONFIG_MVC1_DECODER 0
#define CONFIG_MVC2_DECODER 0
#define CONFIG_MVDV_DECODER 0
#define CONFIG_MVHA_DECODER 0
#define CONFIG_MWSC_DECODER 0
#define CONFIG_MXPEG_DECODER 0
#define CONFIG_NOTCHLC_DECODER 0
#define CONFIG_NUV_DECODER 0
#define CONFIG_PAF_VIDEO_DECODER 0
#define CONFIG_PAM_DECODER 0
#define CONFIG_PBM_DECODER 0
#define CONFIG_PCX_DECODER 0
#define CONFIG_PDV_DECODER 0
#define CONFIG_PFM_DECODER 0
#define CONFIG_PGM_DECODER 0
#define CONFIG_PGMYUV_DECODER 0
#define CONFIG_PGX_DECODER 0
#define CONFIG_PHM_DECODER 0
#define CONFIG_PHOTOCD_DECODER 0
#define CONFIG_PICTOR_DECODER 0
#define CONFIG_PIXLET_DECODER 0
#define CONFIG_PNG_DECODER 0
#define CONFIG_PPM_DECODER 0
#define CONFIG_PRORES_DECODER 0
#define CONFIG_PRORES_RAW_DECODER 0
#define CONFIG_PROSUMER_DECODER 0
#define CONFIG_PSD_DECODER 0
#define CONFIG_PTX_DECODER 0
#define CONFIG_QDRAW_DECODER 0
#define CONFIG_QOI_DECODER 0
#define CONFIG_QPEG_DECODER 0
#define CONFIG_QTRLE_DECODER 0
#define CONFIG_R10K_DECODER 0
#define CONFIG_R210_DECODER 0
#define CONFIG_RASC_DECODER 0
#define CONFIG_RAWVIDEO_DECODER 0
#define CONFIG_RKA_DECODER 0
#define CONFIG_RL2_DECODER 0
#define CONFIG_ROQ_DECODER 0
#define CONFIG_RPZA_DECODER 0
#define CONFIG_RSCC_DECODER 0
#define CONFIG_RTV1_DECODER 0
#define CONFIG_RV10_DECODER 0
#define CONFIG_RV20_DECODER 0
#define CONFIG_RV30_DECODER 0
#define CONFIG_RV40_DECODER 0
#define CONFIG_RV60_DECODER 0
#define CONFIG_S302M_DECODER 0
#define CONFIG_SANM_DECODER 0
#define CONFIG_SCPR_DECODER 0
#define CONFIG_SCREENPRESSO_DECODER 0
#define CONFIG_SGA_DECODER 0
#define CONFIG_SGI_DECODER 0
#define CONFIG_SGIRLE_DECODER 0
#define CONFIG_SHEERVIDEO_DECODER 0
#define CONFIG_SIMBIOSIS_IMX_DECODER 0
#define CONFIG_SMACKER_DECODER 0
#define CONFIG_SMC_DECODER 0
#define CONFIG_SMVJPEG_DECODER 0
#define CONFIG_SNOW_DECODER 0
#define CONFIG_SP5X_DECODER 0
#define CONFIG_SPEEDHQ_DECODER 0
#define CONFIG_SPEEX_DECODER 0
#define CONFIG_SRGC_DECODER 0
#define CONFIG_SUNRAST_DECODER 0
#define CONFIG_SVQ1_DECODER 0
#define CONFIG_SVQ3_DECODER 0
#define CONFIG_TARGA_DECODER 0
#define CONFIG_TARGA_Y216_DECODER 0
#define CONFIG_TDSC_DECODER 0
#define CONFIG_THEORA_DECODER 0
#define CONFIG_THP_DECODER 0
#define CONFIG_TIERTEXSEQVIDEO_DECODER 0
#define CONFIG_TIFF_DECODER 0
#define CONFIG_TMV_DECODER 0
#define CONFIG_TRUEMOTION1_DECODER 0
#define CONFIG_TRUEMOTION2_DECODER 0
#define CONFIG_TRUEMOTION2RT_DECODER 0
#define CONFIG_TSCC_DECODER 0
#define CONFIG_TSCC2_DECODER 0
#define CONFIG_TXD_DECODER 0
#define CONFIG_ULTI_DECODER 0
#define CONFIG_UTVIDEO_DECODER 0
#define CONFIG_V210_DECODER 0
#define CONFIG_V210X_DECODER 0
#define CONFIG_V308_DECODER 0
#define CONFIG_V408_DECODER 0
#define CONFIG_V410_DECODER 0
#define CONFIG_VB_DECODER 0
#define CONFIG_VBN_DECODER 0
#define CONFIG_VBLE_DECODER 0
#define CONFIG_VC1_DECODER 0
#define CONFIG_VC1IMAGE_DECODER 0
#define CONFIG_VC1_MMAL_DECODER 0
#define CONFIG_VC1_QSV_DECODER 0
#define CONFIG_VC1_V4L2M2M_DECODER 0
#define CONFIG_VCR1_DECODER 0
#define CONFIG_VMDVIDEO_DECODER 0
#define CONFIG_VMIX_DECODER 0
#define CONFIG_VMNC_DECODER 0
#define CONFIG_VP3_DECODER 0
#define CONFIG_VP4_DECODER 0
#define CONFIG_VP5_DECODER 0
#define CONFIG_VP6_DECODER 0
#define CONFIG_VP6A_DECODER 0
#define CONFIG_VP6F_DECODER 0
#define CONFIG_VP7_DECODER 0
#define CONFIG_VP8_DECODER 0
#define CONFIG_VP8_RKMPP_DECODER 0
#define CONFIG_VP8_V4L2M2M_DECODER 0
#define CONFIG_VP9_DECODER 0
#define CONFIG_VP9_RKMPP_DECODER 0
#define CONFIG_VP9_V4L2M2M_DECODER 0
#define CONFIG_VQA_DECODER 0
#define CONFIG_VQC_DECODER 0
#define CONFIG_VVC_DECODER 0
#define CONFIG_WBMP_DECODER 0
#define CONFIG_WEBP_DECODER 0
#define CONFIG_WCMV_DECODER 0
#define CONFIG_WRAPPED_AVFRAME_DECODER 0
#define CONFIG_WMV1_DECODER 0
#define CONFIG_WMV2_DECODER 0
#define CONFIG_WMV3_DECODER 0
#define CONFIG_WMV3IMAGE_DECODER 0
#define CONFIG_WNV1_DECODER 0
#define CONFIG_XAN_WC3_DECODER 0
#define CONFIG_XAN_WC4_DECODER 0
#define CONFIG_XBM_DECODER 0
#define CONFIG_XFACE_DECODER 0
#define CONFIG_XL_DECODER 0
#define CONFIG_XPM_DECODER 0
#define CONFIG_XWD_DECODER 0
#define CONFIG_Y41P_DECODER 0
#define CONFIG_YLC_DECODER 0
#define CONFIG_YOP_DECODER 0
#define CONFIG_YUV4_DECODER 0
#define CONFIG_ZERO12V_DECODER 0
#define CONFIG_ZEROCODEC_DECODER 0
#define CONFIG_ZLIB_DECODER 0
#define CONFIG_ZMBV_DECODER 0
#define CONFIG_AAC_DECODER 1
#define CONFIG_AAC_FIXED_DECODER 0
#define CONFIG_AAC_LATM_DECODER 0
#define CONFIG_AC3_DECODER 0
#define CONFIG_AC3_FIXED_DECODER 0
#define CONFIG_ACELP_KELVIN_DECODER 0
#define CONFIG_AHX_DECODER 0
#define CONFIG_ALAC_DECODER 0
#define CONFIG_ALS_DECODER 0
#define CONFIG_AMRNB_DECODER 0
#define CONFIG_AMRWB_DECODER 0
#define CONFIG_APAC_DECODER 0
#define CONFIG_APE_DECODER 0
#define CONFIG_APTX_DECODER 0
#define CONFIG_APTX_HD_DECODER 0
#define CONFIG_ATRAC1_DECODER 0
#define CONFIG_ATRAC3_DECODER 0
#define CONFIG_ATRAC3AL_DECODER 0
#define CONFIG_ATRAC3P_DECODER 0
#define CONFIG_ATRAC3PAL_DECODER 0
#define CONFIG_ATRAC9_DECODER 0
#define CONFIG_BINKAUDIO_DCT_DECODER 0
#define CONFIG_BINKAUDIO_RDFT_DECODER 0
#define CONFIG_BMV_AUDIO_DECODER 0
#define CONFIG_BONK_DECODER 0
#define CONFIG_COOK_DECODER 0
#define CONFIG_DCA_DECODER 0
#define CONFIG_DFPWM_DECODER 0
#define CONFIG_DOLBY_E_DECODER 0
#define CONFIG_DSD_LSBF_DECODER 0
#define CONFIG_DSD_MSBF_DECODER 0
#define CONFIG_DSD_LSBF_PLANAR_DECODER 0
#define CONFIG_DSD_MSBF_PLANAR_DECODER 0
#define CONFIG_DSICINAUDIO_DECODER 0
#define CONFIG_DSS_SP_DECODER 0
#define CONFIG_DST_DECODER 0
#define CONFIG_EAC3_DECODER 0
#define CONFIG_EVRC_DECODER 0
#define CONFIG_FASTAUDIO_DECODER 0
#define CONFIG_FFWAVESYNTH_DECODER 0
#define CONFIG_FLAC_DECODER 1
#define CONFIG_FTR_DECODER 0
#define CONFIG_G723_1_DECODER 0
#define CONFIG_G728_DECODER 0
#define CONFIG_G729_DECODER 0
#define CONFIG_GSM_DECODER 0
#define CONFIG_GSM_MS_DECODER 0
#define CONFIG_HCA_DECODER 0
#define CONFIG_HCOM_DECODER 0
#define CONFIG_HDR_DECODER 0
#define CONFIG_IAC_DECODER 0
#define CONFIG_ILBC_DECODER 0
#define CONFIG_IMC_DECODER 0
#define CONFIG_INTERPLAY_ACM_DECODER 0
#define CONFIG_MACE3_DECODER 0
#define CONFIG_MACE6_DECODER 0
#define CONFIG_METASOUND_DECODER 0
#define CONFIG_MISC4_DECODER 0
#define CONFIG_MLP_DECODER 0
#define CONFIG_MP1_DECODER 0
#define CONFIG_MP1FLOAT_DECODER 0
#define CONFIG_MP2_DECODER 0
#define CONFIG_MP2FLOAT_DECODER 0
#define CONFIG_MP3FLOAT_DECODER 0
#define CONFIG_MP3_DECODER 1
#define CONFIG_MP3ADUFLOAT_DECODER 0
#define CONFIG_MP3ADU_DECODER 0
#define CONFIG_MP3ON4FLOAT_DECODER 0
#define CONFIG_MP3ON4_DECODER 0
#define CONFIG_MPC7_DECODER 0
#define CONFIG_MPC8_DECODER 0
#define CONFIG_MSNSIREN_DECODER 0
#define CONFIG_NELLYMOSER_DECODER 0
#define CONFIG_ON2AVC_DECODER 0
#define CONFIG_OPUS_DECODER 0
#define CONFIG_OSQ_DECODER 0
#define CONFIG_PAF_AUDIO_DECODER 0
#define CONFIG_QCELP_DECODER 0
#define CONFIG_QDM2_DECODER 0
#define CONFIG_QDMC_DECODER 0
#define CONFIG_QOA_DECODER 0
#define CONFIG_RA_144_DECODER 0
#define CONFIG_RA_288_DECODER 0
#define CONFIG_RALF_DECODER 0
#define CONFIG_SBC_DECODER 0
#define CONFIG_SHORTEN_DECODER 0
#define CONFIG_SIPR_DECODER 0
#define CONFIG_SIREN_DECODER 0
#define CONFIG_SMACKAUD_DECODER 0
#define CONFIG_SONIC_DECODER 0
#define CONFIG_TAK_DECODER 0
#define CONFIG_TRUEHD_DECODER 0
#define CONFIG_TRUESPEECH_DECODER 0
#define CONFIG_TTA_DECODER 0
#define CONFIG_TWINVQ_DECODER 0
#define CONFIG_VMDAUDIO_DECODER 0
#define CONFIG_VORBIS_DECODER 1
#define CONFIG_WAVARC_DECODER 0
#define CONFIG_WAVPACK_DECODER 0
#define CONFIG_WMALOSSLESS_DECODER 0
#define CONFIG_WMAPRO_DECODER 0
#define CONFIG_WMAV1_DECODER 0
#define CONFIG_WMAV2_DECODER 0
#define CONFIG_WMAVOICE_DECODER 0
#define CONFIG_WS_SND1_DECODER 0
#define CONFIG_XMA1_DECODER 0
#define CONFIG_XMA2_DECODER 0
#define CONFIG_PCM_ALAW_DECODER 1
#define CONFIG_PCM_BLURAY_DECODER 0
#define CONFIG_PCM_DVD_DECODER 0
#define CONFIG_PCM_F16LE_DECODER 0
#define CONFIG_PCM_F24LE_DECODER 0
#define CONFIG_PCM_F32BE_DECODER 0
#define CONFIG_PCM_F32LE_DECODER 1
#define CONFIG_PCM_F64BE_DECODER 0
#define CONFIG_PCM_F64LE_DECODER 0
#define CONFIG_PCM_LXF_DECODER 0
#define CONFIG_PCM_MULAW_DECODER 1
#define CONFIG_PCM_S8_DECODER 0
#define CONFIG_PCM_S8_PLANAR_DECODER 0
#define CONFIG_PCM_S16BE_DECODER 1
#define CONFIG_PCM_S16BE_PLANAR_DECODER 0
#define CONFIG_PCM_S16LE_DECODER 1
#define CONFIG_PCM_S16LE_PLANAR_DECODER 0
#define CONFIG_PCM_S24BE_DECODER 1
#define CONFIG_PCM_S24DAUD_DECODER 0
#define CONFIG_PCM_S24LE_DECODER 1
#define CONFIG_PCM_S24LE_PLANAR_DECODER 0
#define CONFIG_PCM_S32BE_DECODER 0
#define CONFIG_PCM_S32LE_DECODER 1
#define CONFIG_PCM_S32LE_PLANAR_DECODER 0
#define CONFIG_PCM_S64BE_DECODER 0
#define CONFIG_PCM_S64LE_DECODER 0
#define CONFIG_PCM_SGA_DECODER 0
#define CONFIG_PCM_U8_DECODER 1
#define CONFIG_PCM_U16BE_DECODER 0
#define CONFIG_PCM_U16LE_DECODER 0
#define CONFIG_PCM_U24BE_DECODER 0
#define CONFIG_PCM_U24LE_DECODER 0
#define CONFIG_PCM_U32BE_DECODER 0
#define CONFIG_PCM_U32LE_DECODER 0
#define CONFIG_PCM_VIDC_DECODER 0
#define CONFIG_CBD2_DPCM_DECODER 0
#define CONFIG_DERF_DPCM_DECODER 0
#define CONFIG_GREMLIN_DPCM_DECODER 0
#define CONFIG_INTERPLAY_DPCM_DECODER 0
#define CONFIG_ROQ_DPCM_DECODER 0
#define CONFIG_SDX2_DPCM_DECODER 0
#define CONFIG_SOL_DPCM_DECODER 0
#define CONFIG_XAN_DPCM_DECODER 0
#define CONFIG_WADY_DPCM_DECODER 0
#define CONFIG_ADPCM_4XM_DECODER 0
#define CONFIG_ADPCM_ADX_DECODER 0
#define CONFIG_ADPCM_AFC_DECODER 0
#define CONFIG_ADPCM_AGM_DECODER 0
#define CONFIG_ADPCM_AICA_DECODER 0
#define CONFIG_ADPCM_ARGO_DECODER 0
#define CONFIG_ADPCM_CIRCUS_DECODER 0
#define CONFIG_ADPCM_CT_DECODER 0
#define CONFIG_ADPCM_DTK_DECODER 0
#define CONFIG_ADPCM_EA_DECODER 0
#define CONFIG_ADPCM_EA_MAXIS_XA_DECODER 0
#define CONFIG_ADPCM_EA_R1_DECODER 0
#define CONFIG_ADPCM_EA_R2_DECODER 0
#define CONFIG_ADPCM_EA_R3_DECODER 0
#define CONFIG_ADPCM_EA_XAS_DECODER 0
#define CONFIG_ADPCM_G722_DECODER 0
#define CONFIG_ADPCM_G726_DECODER 0
#define CONFIG_ADPCM_G726LE_DECODER 0
#define CONFIG_ADPCM_IMA_ACORN_DECODER 0
#define CONFIG_ADPCM_IMA_AMV_DECODER 0
#define CONFIG_ADPCM_IMA_ALP_DECODER 0
#define CONFIG_ADPCM_IMA_APC_DECODER 0
#define CONFIG_ADPCM_IMA_APM_DECODER 0
#define CONFIG_ADPCM_IMA_CUNNING_DECODER 0
#define CONFIG_ADPCM_IMA_DAT4_DECODER 0
#define CONFIG_ADPCM_IMA_DK3_DECODER 0
#define CONFIG_ADPCM_IMA_DK4_DECODER 0
#define CONFIG_ADPCM_IMA_EA_EACS_DECODER 0
#define CONFIG_ADPCM_IMA_EA_SEAD_DECODER 0
#define CONFIG_ADPCM_IMA_ESCAPE_DECODER 0
#define CONFIG_ADPCM_IMA_HVQM2_DECODER 0
#define CONFIG_ADPCM_IMA_HVQM4_DECODER 0
#define CONFIG_ADPCM_IMA_ISS_DECODER 0
#define CONFIG_ADPCM_IMA_MAGIX_DECODER 0
#define CONFIG_ADPCM_IMA_MOFLEX_DECODER 0
#define CONFIG_ADPCM_IMA_MTF_DECODER 0
#define CONFIG_ADPCM_IMA_OKI_DECODER 0
#define CONFIG_ADPCM_IMA_PDA_DECODER 0
#define CONFIG_ADPCM_IMA_QT_DECODER 0
#define CONFIG_ADPCM_IMA_RAD_DECODER 0
#define CONFIG_ADPCM_IMA_SSI_DECODER 0
#define CONFIG_ADPCM_IMA_SMJPEG_DECODER 0
#define CONFIG_ADPCM_IMA_WAV_DECODER 0
#define CONFIG_ADPCM_IMA_WS_DECODER 0
#define CONFIG_ADPCM_IMA_XBOX_DECODER 0
#define CONFIG_ADPCM_MS_DECODER 0
#define CONFIG_ADPCM_MTAF_DECODER 0
#define CONFIG_ADPCM_N64_DECODER 0
#define CONFIG_ADPCM_PSX_DECODER 0
#define CONFIG_ADPCM_PSXC_DECODER 0
#define CONFIG_ADPCM_SANYO_DECODER 0
#define CONFIG_ADPCM_SBPRO_2_DECODER 0
#define CONFIG_ADPCM_SBPRO_3_DECODER 0
#define CONFIG_ADPCM_SBPRO_4_DECODER 0
#define CONFIG_ADPCM_SWF_DECODER 0
#define CONFIG_ADPCM_THP_DECODER 0
#define CONFIG_ADPCM_THP_LE_DECODER 0
#define CONFIG_ADPCM_VIMA_DECODER 0
#define CONFIG_ADPCM_XA_DECODER 0
#define CONFIG_ADPCM_XMD_DECODER 0
#define CONFIG_ADPCM_YAMAHA_DECODER 0
#define CONFIG_ADPCM_ZORK_DECODER 0
#define CONFIG_SSA_DECODER 0
#define CONFIG_ASS_DECODER 0
#define CONFIG_CCAPTION_DECODER 0
#define CONFIG_DVBSUB_DECODER 0
#define CONFIG_DVDSUB_DECODER 0
#define CONFIG_JACOSUB_DECODER 0
#define CONFIG_MICRODVD_DECODER 0
#define CONFIG_MOVTEXT_DECODER 0
#define CONFIG_MPL2_DECODER 0
#define CONFIG_PGSSUB_DECODER 0
#define CONFIG_PJS_DECODER 0
#define CONFIG_REALTEXT_DECODER 0
#define CONFIG_SAMI_DECODER 0
#define CONFIG_SRT_DECODER 0
#define CONFIG_STL_DECODER 0
#define CONFIG_SUBRIP_DECODER 0
#define CONFIG_SUBVIEWER_DECODER 0
#define CONFIG_SUBVIEWER1_DECODER 0
#define CONFIG_TEXT_DECODER 0
#define CONFIG_VPLAYER_DECODER 0
#define CONFIG_WEBVTT_DECODER 0
#define CONFIG_XSUB_DECODER 0
#define CONFIG_AAC_AT_DECODER 0
#define CONFIG_AC3_AT_DECODER 0
#define CONFIG_ADPCM_IMA_QT_AT_DECODER 0
#define CONFIG_ALAC_AT_DECODER 0
#define CONFIG_AMR_NB_AT_DECODER 0
#define CONFIG_EAC3_AT_DECODER 0
#define CONFIG_GSM_MS_AT_DECODER 0
#define CONFIG_ILBC_AT_DECODER 0
#define CONFIG_MP1_AT_DECODER 0
#define CONFIG_MP2_AT_DECODER 0
#define CONFIG_MP3_AT_DECODER 0
#define CONFIG_PCM_ALAW_AT_DECODER 0
#define CONFIG_PCM_MULAW_AT_DECODER 0
#define CONFIG_QDMC_AT_DECODER 0
#define CONFIG_QDM2_AT_DECODER 0
#define CONFIG_LIBARIBCAPTION_DECODER 0
#define CONFIG_LIBARIBB24_DECODER 0
#define CONFIG_LIBCELT_DECODER 0
#define CONFIG_LIBCODEC2_DECODER 0
#define CONFIG_LIBDAV1D_DECODER 0
#define CONFIG_LIBDAVS2_DECODER 0
#define CONFIG_LIBFDK_AAC_DECODER 0
#define CONFIG_LIBGSM_DECODER 0
#define CONFIG_LIBGSM_MS_DECODER 0
#define CONFIG_LIBILBC_DECODER 0
#define CONFIG_LIBJXL_ANIM_DECODER 0
#define CONFIG_LIBJXL_DECODER 0
#define CONFIG_LIBLC3_DECODER 0
#define CONFIG_LIBMPEGHDEC_DECODER 0
#define CONFIG_LIBOPENCORE_AMRNB_DECODER 0
#define CONFIG_LIBOPENCORE_AMRWB_DECODER 0
#define CONFIG_LIBOPUS_DECODER 1
#define CONFIG_LIBRSVG_DECODER 0
#define CONFIG_LIBSPEEX_DECODER 0
#define CONFIG_LIBSVTJPEGXS_DECODER 0
#define CONFIG_LIBUAVS3D_DECODER 0
#define CONFIG_LIBVORBIS_DECODER 0
#define CONFIG_LIBVPX_VP8_DECODER 0
#define CONFIG_LIBVPX_VP9_DECODER 0
#define CONFIG_LIBXEVD_DECODER 0
#define CONFIG_LIBZVBI_TELETEXT_DECODER 0
#define CONFIG_BINTEXT_DECODER 0
#define CONFIG_XBIN_DECODER 0
#define CONFIG_IDF_DECODER 0
#define CONFIG_AAC_MEDIACODEC_DECODER 0
#define CONFIG_AMRNB_MEDIACODEC_DECODER 0
#define CONFIG_AMRWB_MEDIACODEC_DECODER 0
#define CONFIG_LIBAOM_AV1_DECODER 0
#define CONFIG_AV1_DECODER 0
#define CONFIG_AV1_CUVID_DECODER 0
#define CONFIG_AV1_MEDIACODEC_DECODER 0
#define CONFIG_AV1_QSV_DECODER 0
#define CONFIG_AV1_AMF_DECODER 0
#define CONFIG_LIBOPENH264_DECODER 0
#define CONFIG_H264_AMF_DECODER 0
#define CONFIG_H264_CUVID_DECODER 0
#define CONFIG_H264_OH_DECODER 0
#define CONFIG_HEVC_AMF_DECODER 0
#define CONFIG_HEVC_CUVID_DECODER 0
#define CONFIG_HEVC_MEDIACODEC_DECODER 0
#define CONFIG_HEVC_OH_DECODER 0
#define CONFIG_MJPEG_CUVID_DECODER 0
#define CONFIG_MJPEG_QSV_DECODER 0
#define CONFIG_MP3_MEDIACODEC_DECODER 0
#define CONFIG_MPEG1_CUVID_DECODER 0
#define CONFIG_MPEG2_CUVID_DECODER 0
#define CONFIG_MPEG4_CUVID_DECODER 0
#define CONFIG_MPEG4_MEDIACODEC_DECODER 0
#define CONFIG_VC1_CUVID_DECODER 0
#define CONFIG_VP8_CUVID_DECODER 0
#define CONFIG_VP8_MEDIACODEC_DECODER 0
#define CONFIG_VP8_QSV_DECODER 0
#define CONFIG_VP9_AMF_DECODER 0
#define CONFIG_VP9_CUVID_DECODER 0
#define CONFIG_VP9_MEDIACODEC_DECODER 0
#define CONFIG_VP9_QSV_DECODER 0
#define CONFIG_VVC_QSV_DECODER 0
#define CONFIG_VNULL_DECODER 0
#define CONFIG_ANULL_DECODER 0
#define CONFIG_A64MULTI_ENCODER 0
#define CONFIG_A64MULTI5_ENCODER 0
#define CONFIG_ALIAS_PIX_ENCODER 0
#define CONFIG_AMV_ENCODER 0
#define CONFIG_APNG_ENCODER 0
#define CONFIG_ASV1_ENCODER 0
#define CONFIG_ASV2_ENCODER 0
#define CONFIG_AVRP_ENCODER 0
#define CONFIG_AVUI_ENCODER 0
#define CONFIG_BITPACKED_ENCODER 0
#define CONFIG_BMP_ENCODER 0
#define CONFIG_CFHD_ENCODER 0
#define CONFIG_CINEPAK_ENCODER 0
#define CONFIG_CLJR_ENCODER 0
#define CONFIG_COMFORTNOISE_ENCODER 0
#define CONFIG_DNXHD_ENCODER 0
#define CONFIG_DPX_ENCODER 0
#define CONFIG_DVVIDEO_ENCODER 0
#define CONFIG_DXV_ENCODER 0
#define CONFIG_EXR_ENCODER 0
#define CONFIG_FFV1_ENCODER 0
#define CONFIG_FFV1_VULKAN_ENCODER 0
#define CONFIG_FFVHUFF_ENCODER 0
#define CONFIG_FITS_ENCODER 0
#define CONFIG_FLASHSV_ENCODER 0
#define CONFIG_FLASHSV2_ENCODER 0
#define CONFIG_FLV_ENCODER 0
#define CONFIG_GIF_ENCODER 0
#define CONFIG_H261_ENCODER 0
#define CONFIG_H263_ENCODER 0
#define CONFIG_H263P_ENCODER 0
#define CONFIG_H264_MEDIACODEC_ENCODER 0
#define CONFIG_H264_RKMPP_ENCODER 0
#define CONFIG_HAP_ENCODER 0
#define CONFIG_HEVC_RKMPP_ENCODER 0
#define CONFIG_HUFFYUV_ENCODER 0
#define CONFIG_JPEG2000_ENCODER 0
#define CONFIG_JPEGLS_ENCODER 0
#define CONFIG_LJPEG_ENCODER 0
#define CONFIG_MAGICYUV_ENCODER 0
#define CONFIG_MJPEG_ENCODER 0
#define CONFIG_MPEG1VIDEO_ENCODER 0
#define CONFIG_MPEG2VIDEO_ENCODER 0
#define CONFIG_MPEG4_ENCODER 0
#define CONFIG_MSMPEG4V2_ENCODER 0
#define CONFIG_MSMPEG4V3_ENCODER 0
#define CONFIG_MSRLE_ENCODER 0
#define CONFIG_MSVIDEO1_ENCODER 0
#define CONFIG_PAM_ENCODER 0
#define CONFIG_PBM_ENCODER 0
#define CONFIG_PCX_ENCODER 0
#define CONFIG_PFM_ENCODER 0
#define CONFIG_PGM_ENCODER 0
#define CONFIG_PGMYUV_ENCODER 0
#define CONFIG_PHM_ENCODER 0
#define CONFIG_PNG_ENCODER 0
#define CONFIG_PPM_ENCODER 0
#define CONFIG_PRORES_ENCODER 0
#define CONFIG_PRORES_AW_ENCODER 0
#define CONFIG_PRORES_KS_ENCODER 0
#define CONFIG_PRORES_KS_VULKAN_ENCODER 0
#define CONFIG_QOI_ENCODER 0
#define CONFIG_QTRLE_ENCODER 0
#define CONFIG_R10K_ENCODER 0
#define CONFIG_R210_ENCODER 0
#define CONFIG_RAWVIDEO_ENCODER 0
#define CONFIG_ROQ_ENCODER 0
#define CONFIG_RPZA_ENCODER 0
#define CONFIG_RV10_ENCODER 0
#define CONFIG_RV20_ENCODER 0
#define CONFIG_S302M_ENCODER 0
#define CONFIG_SGI_ENCODER 0
#define CONFIG_SMC_ENCODER 0
#define CONFIG_SNOW_ENCODER 0
#define CONFIG_SPEEDHQ_ENCODER 0
#define CONFIG_SUNRAST_ENCODER 0
#define CONFIG_SVQ1_ENCODER 0
#define CONFIG_TARGA_ENCODER 0
#define CONFIG_TIFF_ENCODER 0
#define CONFIG_UTVIDEO_ENCODER 0
#define CONFIG_V210_ENCODER 0
#define CONFIG_V308_ENCODER 0
#define CONFIG_V408_ENCODER 0
#define CONFIG_V410_ENCODER 0
#define CONFIG_VBN_ENCODER 0
#define CONFIG_VC2_ENCODER 0
#define CONFIG_WBMP_ENCODER 0
#define CONFIG_WRAPPED_AVFRAME_ENCODER 0
#define CONFIG_WMV1_ENCODER 0
#define CONFIG_WMV2_ENCODER 0
#define CONFIG_XBM_ENCODER 0
#define CONFIG_XFACE_ENCODER 0
#define CONFIG_XWD_ENCODER 0
#define CONFIG_Y41P_ENCODER 0
#define CONFIG_YUV4_ENCODER 0
#define CONFIG_ZLIB_ENCODER 0
#define CONFIG_ZMBV_ENCODER 0
#define CONFIG_AAC_ENCODER 0
#define CONFIG_AC3_ENCODER 0
#define CONFIG_AC3_FIXED_ENCODER 0
#define CONFIG_ALAC_ENCODER 0
#define CONFIG_APTX_ENCODER 0
#define CONFIG_APTX_HD_ENCODER 0
#define CONFIG_DCA_ENCODER 0
#define CONFIG_DFPWM_ENCODER 0
#define CONFIG_EAC3_ENCODER 0
#define CONFIG_FLAC_ENCODER 0
#define CONFIG_G723_1_ENCODER 0
#define CONFIG_HDR_ENCODER 0
#define CONFIG_MLP_ENCODER 0
#define CONFIG_MP2_ENCODER 0
#define CONFIG_MP2FIXED_ENCODER 0
#define CONFIG_NELLYMOSER_ENCODER 0
#define CONFIG_OPUS_ENCODER 0
#define CONFIG_RA_144_ENCODER 0
#define CONFIG_SBC_ENCODER 0
#define CONFIG_SONIC_ENCODER 0
#define CONFIG_SONIC_LS_ENCODER 0
#define CONFIG_TRUEHD_ENCODER 0
#define CONFIG_TTA_ENCODER 0
#define CONFIG_VORBIS_ENCODER 0
#define CONFIG_WAVPACK_ENCODER 0
#define CONFIG_WMAV1_ENCODER 0
#define CONFIG_WMAV2_ENCODER 0
#define CONFIG_PCM_ALAW_ENCODER 0
#define CONFIG_PCM_BLURAY_ENCODER 0
#define CONFIG_PCM_DVD_ENCODER 0
#define CONFIG_PCM_F32BE_ENCODER 0
#define CONFIG_PCM_F32LE_ENCODER 0
#define CONFIG_PCM_F64BE_ENCODER 0
#define CONFIG_PCM_F64LE_ENCODER 0
#define CONFIG_PCM_MULAW_ENCODER 0
#define CONFIG_PCM_S8_ENCODER 0
#define CONFIG_PCM_S8_PLANAR_ENCODER 0
#define CONFIG_PCM_S16BE_ENCODER 0
#define CONFIG_PCM_S16BE_PLANAR_ENCODER 0
#define CONFIG_PCM_S16LE_ENCODER 0
#define CONFIG_PCM_S16LE_PLANAR_ENCODER 0
#define CONFIG_PCM_S24BE_ENCODER 0
#define CONFIG_PCM_S24DAUD_ENCODER 0
#define CONFIG_PCM_S24LE_ENCODER 0
#define CONFIG_PCM_S24LE_PLANAR_ENCODER 0
#define CONFIG_PCM_S32BE_ENCODER 0
#define CONFIG_PCM_S32LE_ENCODER 0
#define CONFIG_PCM_S32LE_PLANAR_ENCODER 0
#define CONFIG_PCM_S64BE_ENCODER 0
#define CONFIG_PCM_S64LE_ENCODER 0
#define CONFIG_PCM_U8_ENCODER 0
#define CONFIG_PCM_U16BE_ENCODER 0
#define CONFIG_PCM_U16LE_ENCODER 0
#define CONFIG_PCM_U24BE_ENCODER 0
#define CONFIG_PCM_U24LE_ENCODER 0
#define CONFIG_PCM_U32BE_ENCODER 0
#define CONFIG_PCM_U32LE_ENCODER 0
#define CONFIG_PCM_VIDC_ENCODER 0
#define CONFIG_ROQ_DPCM_ENCODER 0
#define CONFIG_ADPCM_ADX_ENCODER 0
#define CONFIG_ADPCM_ARGO_ENCODER 0
#define CONFIG_ADPCM_G722_ENCODER 0
#define CONFIG_ADPCM_G726_ENCODER 0
#define CONFIG_ADPCM_G726LE_ENCODER 0
#define CONFIG_ADPCM_IMA_AMV_ENCODER 0
#define CONFIG_ADPCM_IMA_ALP_ENCODER 0
#define CONFIG_ADPCM_IMA_APM_ENCODER 0
#define CONFIG_ADPCM_IMA_QT_ENCODER 0
#define CONFIG_ADPCM_IMA_SSI_ENCODER 0
#define CONFIG_ADPCM_IMA_WAV_ENCODER 0
#define CONFIG_ADPCM_IMA_WS_ENCODER 0
#define CONFIG_ADPCM_MS_ENCODER 0
#define CONFIG_ADPCM_SWF_ENCODER 0
#define CONFIG_ADPCM_YAMAHA_ENCODER 0
#define CONFIG_SSA_ENCODER 0
#define CONFIG_ASS_ENCODER 0
#define CONFIG_DVBSUB_ENCODER 0
#define CONFIG_DVDSUB_ENCODER 0
#define CONFIG_MOVTEXT_ENCODER 0
#define CONFIG_SRT_ENCODER 0
#define CONFIG_SUBRIP_ENCODER 0
#define CONFIG_TEXT_ENCODER 0
#define CONFIG_TTML_ENCODER 0
#define CONFIG_WEBVTT_ENCODER 0
#define CONFIG_XSUB_ENCODER 0
#define CONFIG_AAC_AT_ENCODER 0
#define CONFIG_ALAC_AT_ENCODER 0
#define CONFIG_ILBC_AT_ENCODER 0
#define CONFIG_PCM_ALAW_AT_ENCODER 0
#define CONFIG_PCM_MULAW_AT_ENCODER 0
#define CONFIG_LIBAOM_AV1_ENCODER 0
#define CONFIG_LIBCODEC2_ENCODER 0
#define CONFIG_LIBFDK_AAC_ENCODER 0
#define CONFIG_LIBGSM_ENCODER 0
#define CONFIG_LIBGSM_MS_ENCODER 0
#define CONFIG_LIBILBC_ENCODER 0
#define CONFIG_LIBJXL_ANIM_ENCODER 0
#define CONFIG_LIBJXL_ENCODER 0
#define CONFIG_LIBLC3_ENCODER 0
#define CONFIG_LIBMP3LAME_ENCODER 0
#define CONFIG_LIBOAPV_ENCODER 0
#define CONFIG_LIBOPENCORE_AMRNB_ENCODER 0
#define CONFIG_LIBOPENJPEG_ENCODER 0
#define CONFIG_LIBOPUS_ENCODER 0
#define CONFIG_LIBRAV1E_ENCODER 0
#define CONFIG_LIBSHINE_ENCODER 0
#define CONFIG_LIBSPEEX_ENCODER 0
#define CONFIG_LIBSVTAV1_ENCODER 0
#define CONFIG_LIBSVTJPEGXS_ENCODER 0
#define CONFIG_LIBTHEORA_ENCODER 0
#define CONFIG_LIBTWOLAME_ENCODER 0
#define CONFIG_LIBVO_AMRWBENC_ENCODER 0
#define CONFIG_LIBVORBIS_ENCODER 0
#define CONFIG_LIBVPX_VP8_ENCODER 0
#define CONFIG_LIBVPX_VP9_ENCODER 0
#define CONFIG_LIBVVENC_ENCODER 0
#define CONFIG_LIBWEBP_ANIM_ENCODER 0
#define CONFIG_LIBWEBP_ENCODER 0
#define CONFIG_LIBX262_ENCODER 0
#define CONFIG_LIBX264_ENCODER 0
#define CONFIG_LIBX264RGB_ENCODER 0
#define CONFIG_LIBX265_ENCODER 0
#define CONFIG_LIBXEVE_ENCODER 0
#define CONFIG_LIBXAVS_ENCODER 0
#define CONFIG_LIBXAVS2_ENCODER 0
#define CONFIG_LIBXVID_ENCODER 0
#define CONFIG_AAC_MF_ENCODER 0
#define CONFIG_AC3_MF_ENCODER 0
#define CONFIG_H263_V4L2M2M_ENCODER 0
#define CONFIG_AV1_D3D12VA_ENCODER 0
#define CONFIG_AV1_MEDIACODEC_ENCODER 0
#define CONFIG_AV1_NVENC_ENCODER 0
#define CONFIG_AV1_QSV_ENCODER 0
#define CONFIG_AV1_AMF_ENCODER 0
#define CONFIG_AV1_MF_ENCODER 0
#define CONFIG_AV1_VAAPI_ENCODER 0
#define CONFIG_AV1_VULKAN_ENCODER 0
#define CONFIG_LIBOPENH264_ENCODER 0
#define CONFIG_H264_AMF_ENCODER 0
#define CONFIG_H264_D3D12VA_ENCODER 0
#define CONFIG_H264_MF_ENCODER 0
#define CONFIG_H264_NVENC_ENCODER 0
#define CONFIG_H264_OH_ENCODER 0
#define CONFIG_H264_OMX_ENCODER 0
#define CONFIG_H264_QSV_ENCODER 0
#define CONFIG_H264_V4L2M2M_ENCODER 0
#define CONFIG_H264_VAAPI_ENCODER 0
#define CONFIG_H264_VIDEOTOOLBOX_ENCODER 0
#define CONFIG_H264_VULKAN_ENCODER 0
#define CONFIG_HEVC_AMF_ENCODER 0
#define CONFIG_HEVC_D3D12VA_ENCODER 0
#define CONFIG_HEVC_MEDIACODEC_ENCODER 0
#define CONFIG_HEVC_MF_ENCODER 0
#define CONFIG_HEVC_NVENC_ENCODER 0
#define CONFIG_HEVC_OH_ENCODER 0
#define CONFIG_HEVC_QSV_ENCODER 0
#define CONFIG_HEVC_V4L2M2M_ENCODER 0
#define CONFIG_HEVC_VAAPI_ENCODER 0
#define CONFIG_HEVC_VIDEOTOOLBOX_ENCODER 0
#define CONFIG_HEVC_VULKAN_ENCODER 0
#define CONFIG_LIBKVAZAAR_ENCODER 0
#define CONFIG_MJPEG_QSV_ENCODER 0
#define CONFIG_MJPEG_VAAPI_ENCODER 0
#define CONFIG_MP3_MF_ENCODER 0
#define CONFIG_MPEG2_QSV_ENCODER 0
#define CONFIG_MPEG2_VAAPI_ENCODER 0
#define CONFIG_MPEG4_MEDIACODEC_ENCODER 0
#define CONFIG_MPEG4_OMX_ENCODER 0
#define CONFIG_MPEG4_V4L2M2M_ENCODER 0
#define CONFIG_PRORES_VIDEOTOOLBOX_ENCODER 0
#define CONFIG_VP8_MEDIACODEC_ENCODER 0
#define CONFIG_VP8_V4L2M2M_ENCODER 0
#define CONFIG_VP8_VAAPI_ENCODER 0
#define CONFIG_VP9_MEDIACODEC_ENCODER 0
#define CONFIG_VP9_VAAPI_ENCODER 0
#define CONFIG_VP9_QSV_ENCODER 0
#define CONFIG_VNULL_ENCODER 0
#define CONFIG_ANULL_ENCODER 0
#define CONFIG_AV1_D3D11VA_HWACCEL 0
#define CONFIG_AV1_D3D11VA2_HWACCEL 0
#define CONFIG_AV1_D3D12VA_HWACCEL 0
#define CONFIG_AV1_DXVA2_HWACCEL 0
#define CONFIG_AV1_NVDEC_HWACCEL 0
#define CONFIG_AV1_VAAPI_HWACCEL 0
#define CONFIG_AV1_VDPAU_HWACCEL 0
#define CONFIG_AV1_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_AV1_VULKAN_HWACCEL 0
#define CONFIG_DPX_VULKAN_HWACCEL 0
#define CONFIG_FFV1_VULKAN_HWACCEL 0
#define CONFIG_H263_VAAPI_HWACCEL 0
#define CONFIG_H263_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_H264_D3D11VA_HWACCEL 0
#define CONFIG_H264_D3D11VA2_HWACCEL 0
#define CONFIG_H264_D3D12VA_HWACCEL 0
#define CONFIG_H264_DXVA2_HWACCEL 0
#define CONFIG_H264_NVDEC_HWACCEL 0
#define CONFIG_H264_VAAPI_HWACCEL 0
#define CONFIG_H264_VDPAU_HWACCEL 0
#define CONFIG_H264_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_H264_VULKAN_HWACCEL 0
#define CONFIG_HEVC_D3D11VA_HWACCEL 0
#define CONFIG_HEVC_D3D11VA2_HWACCEL 0
#define CONFIG_HEVC_D3D12VA_HWACCEL 0
#define CONFIG_HEVC_DXVA2_HWACCEL 0
#define CONFIG_HEVC_NVDEC_HWACCEL 0
#define CONFIG_HEVC_VAAPI_HWACCEL 0
#define CONFIG_HEVC_VDPAU_HWACCEL 0
#define CONFIG_HEVC_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_HEVC_VULKAN_HWACCEL 0
#define CONFIG_MJPEG_NVDEC_HWACCEL 0
#define CONFIG_MJPEG_VAAPI_HWACCEL 0
#define CONFIG_MPEG1_NVDEC_HWACCEL 0
#define CONFIG_MPEG1_VDPAU_HWACCEL 0
#define CONFIG_MPEG1_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_MPEG2_D3D11VA_HWACCEL 0
#define CONFIG_MPEG2_D3D11VA2_HWACCEL 0
#define CONFIG_MPEG2_D3D12VA_HWACCEL 0
#define CONFIG_MPEG2_DXVA2_HWACCEL 0
#define CONFIG_MPEG2_NVDEC_HWACCEL 0
#define CONFIG_MPEG2_VAAPI_HWACCEL 0
#define CONFIG_MPEG2_VDPAU_HWACCEL 0
#define CONFIG_MPEG2_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_MPEG4_NVDEC_HWACCEL 0
#define CONFIG_MPEG4_VAAPI_HWACCEL 0
#define CONFIG_MPEG4_VDPAU_HWACCEL 0
#define CONFIG_MPEG4_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_PRORES_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_PRORES_VULKAN_HWACCEL 0
#define CONFIG_PRORES_RAW_VULKAN_HWACCEL 0
#define CONFIG_VC1_D3D11VA_HWACCEL 0
#define CONFIG_VC1_D3D11VA2_HWACCEL 0
#define CONFIG_VC1_D3D12VA_HWACCEL 0
#define CONFIG_VC1_DXVA2_HWACCEL 0
#define CONFIG_VC1_NVDEC_HWACCEL 0
#define CONFIG_VC1_VAAPI_HWACCEL 0
#define CONFIG_VC1_VDPAU_HWACCEL 0
#define CONFIG_VP8_NVDEC_HWACCEL 0
#define CONFIG_VP8_VAAPI_HWACCEL 0
#define CONFIG_VP9_D3D11VA_HWACCEL 0
#define CONFIG_VP9_D3D11VA2_HWACCEL 0
#define CONFIG_VP9_D3D12VA_HWACCEL 0
#define CONFIG_VP9_DXVA2_HWACCEL 0
#define CONFIG_VP9_NVDEC_HWACCEL 0
#define CONFIG_VP9_VAAPI_HWACCEL 0
#define CONFIG_VP9_VDPAU_HWACCEL 0
#define CONFIG_VP9_VIDEOTOOLBOX_HWACCEL 0
#define CONFIG_VP9_VULKAN_HWACCEL 0
#define CONFIG_VVC_VAAPI_HWACCEL 0
#define CONFIG_WMV3_D3D11VA_HWACCEL 0
#define CONFIG_WMV3_D3D11VA2_HWACCEL 0
#define CONFIG_WMV3_D3D12VA_HWACCEL 0
#define CONFIG_WMV3_DXVA2_HWACCEL 0
#define CONFIG_WMV3_NVDEC_HWACCEL 0
#define CONFIG_WMV3_VAAPI_HWACCEL 0
#define CONFIG_WMV3_VDPAU_HWACCEL 0
#define CONFIG_AAC_PARSER 1
#define CONFIG_AAC_LATM_PARSER 0
#define CONFIG_AC3_PARSER 0
#define CONFIG_ADX_PARSER 0
#define CONFIG_AHX_PARSER 0
#define CONFIG_AMR_PARSER 0
#define CONFIG_APV_PARSER 0
#define CONFIG_AV1_PARSER 0
#define CONFIG_AVS2_PARSER 0
#define CONFIG_AVS3_PARSER 0
#define CONFIG_BMP_PARSER 0
#define CONFIG_CAVSVIDEO_PARSER 0
#define CONFIG_COOK_PARSER 0
#define CONFIG_CRI_PARSER 0
#define CONFIG_DCA_PARSER 0
#define CONFIG_DIRAC_PARSER 0
#define CONFIG_DNXHD_PARSER 0
#define CONFIG_DNXUC_PARSER 0
#define CONFIG_DOLBY_E_PARSER 0
#define CONFIG_DPX_PARSER 0
#define CONFIG_DVAUDIO_PARSER 0
#define CONFIG_DVBSUB_PARSER 0
#define CONFIG_DVDSUB_PARSER 0
#define CONFIG_DVD_NAV_PARSER 0
#define CONFIG_EVC_PARSER 0
#define CONFIG_FLAC_PARSER 1
#define CONFIG_FTR_PARSER 0
#define CONFIG_FFV1_PARSER 0
#define CONFIG_G723_1_PARSER 0
#define CONFIG_G729_PARSER 0
#define CONFIG_GIF_PARSER 0
#define CONFIG_GSM_PARSER 0
#define CONFIG_H261_PARSER 0
#define CONFIG_H263_PARSER 0
#define CONFIG_H264_PARSER 0
#define CONFIG_HEVC_PARSER 0
#define CONFIG_HDR_PARSER 0
#define CONFIG_IPU_PARSER 0
#define CONFIG_JPEG2000_PARSER 0
#define CONFIG_JPEGXL_PARSER 0
#define CONFIG_JPEGXS_PARSER 0
#define CONFIG_LCEVC_PARSER 0
#define CONFIG_MISC4_PARSER 0
#define CONFIG_MJPEG_PARSER 0
#define CONFIG_MLP_PARSER 0
#define CONFIG_MPEG4VIDEO_PARSER 0
#define CONFIG_MPEGAUDIO_PARSER 1
#define CONFIG_MPEGVIDEO_PARSER 0
#define CONFIG_OPUS_PARSER 1
#define CONFIG_PRORES_PARSER 0
#define CONFIG_PNG_PARSER 0
#define CONFIG_PNM_PARSER 0
#define CONFIG_PRORES_RAW_PARSER 0
#define CONFIG_QOI_PARSER 0
#define CONFIG_RV34_PARSER 0
#define CONFIG_SBC_PARSER 0
#define CONFIG_SIPR_PARSER 0
#define CONFIG_TAK_PARSER 0
#define CONFIG_VC1_PARSER 0
#define CONFIG_VORBIS_PARSER 1
#define CONFIG_VP3_PARSER 0
#define CONFIG_VP8_PARSER 0
#define CONFIG_VP9_PARSER 1
#define CONFIG_VVC_PARSER 0
#define CONFIG_WEBP_PARSER 0
#define CONFIG_XBM_PARSER 0
#define CONFIG_XMA_PARSER 0
#define CONFIG_XWD_PARSER 0
#define CONFIG_ALSA_INDEV 0
#define CONFIG_ANDROID_CAMERA_INDEV 0
#define CONFIG_AVFOUNDATION_INDEV 0
#define CONFIG_DECKLINK_INDEV 0
#define CONFIG_DSHOW_INDEV 0
#define CONFIG_FBDEV_INDEV 0
#define CONFIG_GDIGRAB_INDEV 0
#define CONFIG_IEC61883_INDEV 0
#define CONFIG_JACK_INDEV 0
#define CONFIG_KMSGRAB_INDEV 0
#define CONFIG_LAVFI_INDEV 0
#define CONFIG_OPENAL_INDEV 0
#define CONFIG_OSS_INDEV 0
#define CONFIG_PULSE_INDEV 0
#define CONFIG_SNDIO_INDEV 0
#define CONFIG_V4L2_INDEV 0
#define CONFIG_VFWCAP_INDEV 0
#define CONFIG_XCBGRAB_INDEV 0
#define CONFIG_LIBCDIO_INDEV 0
#define CONFIG_LIBDC1394_INDEV 0
#define CONFIG_ALSA_OUTDEV 0
#define CONFIG_AUDIOTOOLBOX_OUTDEV 0
#define CONFIG_CACA_OUTDEV 0
#define CONFIG_DECKLINK_OUTDEV 0
#define CONFIG_FBDEV_OUTDEV 0
#define CONFIG_OSS_OUTDEV 0
#define CONFIG_PULSE_OUTDEV 0
#define CONFIG_SNDIO_OUTDEV 0
#define CONFIG_V4L2_OUTDEV 0
#define CONFIG_XV_OUTDEV 0
#define CONFIG_AAP_FILTER 0
#define CONFIG_ABENCH_FILTER 0
#define CONFIG_ACOMPRESSOR_FILTER 0
#define CONFIG_ACONTRAST_FILTER 0
#define CONFIG_ACOPY_FILTER 0
#define CONFIG_ACUE_FILTER 0
#define CONFIG_ACROSSFADE_FILTER 0
#define CONFIG_ACROSSOVER_FILTER 0
#define CONFIG_ACRUSHER_FILTER 0
#define CONFIG_ADECLICK_FILTER 0
#define CONFIG_ADECLIP_FILTER 0
#define CONFIG_ADECORRELATE_FILTER 0
#define CONFIG_ADELAY_FILTER 0
#define CONFIG_ADENORM_FILTER 0
#define CONFIG_ADERIVATIVE_FILTER 0
#define CONFIG_ADRC_FILTER 0
#define CONFIG_ADYNAMICEQUALIZER_FILTER 0
#define CONFIG_ADYNAMICSMOOTH_FILTER 0
#define CONFIG_AECHO_FILTER 0
#define CONFIG_AEMPHASIS_FILTER 0
#define CONFIG_AEVAL_FILTER 0
#define CONFIG_AEXCITER_FILTER 0
#define CONFIG_AFADE_FILTER 0
#define CONFIG_AFFTDN_FILTER 0
#define CONFIG_AFFTFILT_FILTER 0
#define CONFIG_AFIR_FILTER 0
#define CONFIG_AFORMAT_FILTER 0
#define CONFIG_AFREQSHIFT_FILTER 0
#define CONFIG_AFWTDN_FILTER 0
#define CONFIG_AGATE_FILTER 0
#define CONFIG_AIIR_FILTER 0
#define CONFIG_AINTEGRAL_FILTER 0
#define CONFIG_AINTERLEAVE_FILTER 0
#define CONFIG_ALATENCY_FILTER 0
#define CONFIG_ALIMITER_FILTER 0
#define CONFIG_ALLPASS_FILTER 0
#define CONFIG_ALOOP_FILTER 0
#define CONFIG_AMERGE_FILTER 0
#define CONFIG_AMETADATA_FILTER 0
#define CONFIG_AMIX_FILTER 0
#define CONFIG_AMULTIPLY_FILTER 0
#define CONFIG_ANEQUALIZER_FILTER 0
#define CONFIG_ANLMDN_FILTER 0
#define CONFIG_ANLMF_FILTER 0
#define CONFIG_ANLMS_FILTER 0
#define CONFIG_ANULL_FILTER 0
#define CONFIG_APAD_FILTER 0
#define CONFIG_APERMS_FILTER 0
#define CONFIG_APHASER_FILTER 0
#define CONFIG_APHASESHIFT_FILTER 0
#define CONFIG_APSNR_FILTER 0
#define CONFIG_APSYCLIP_FILTER 0
#define CONFIG_APULSATOR_FILTER 0
#define CONFIG_AREALTIME_FILTER 0
#define CONFIG_ARESAMPLE_FILTER 0
#define CONFIG_AREVERSE_FILTER 0
#define CONFIG_ARLS_FILTER 0
#define CONFIG_ARNNDN_FILTER 0
#define CONFIG_ASDR_FILTER 0
#define CONFIG_ASEGMENT_FILTER 0
#define CONFIG_ASELECT_FILTER 0
#define CONFIG_ASENDCMD_FILTER 0
#define CONFIG_ASETNSAMPLES_FILTER 0
#define CONFIG_ASETPTS_FILTER 0
#define CONFIG_ASETRATE_FILTER 0
#define CONFIG_ASETTB_FILTER 0
#define CONFIG_ASHOWINFO_FILTER 0
#define CONFIG_ASIDEDATA_FILTER 0
#define CONFIG_ASISDR_FILTER 0
#define CONFIG_ASOFTCLIP_FILTER 0
#define CONFIG_ASPECTRALSTATS_FILTER 0
#define CONFIG_ASPLIT_FILTER 0
#define CONFIG_ASR_FILTER 0
#define CONFIG_ASTATS_FILTER 0
#define CONFIG_ASTREAMSELECT_FILTER 0
#define CONFIG_ASUBBOOST_FILTER 0
#define CONFIG_ASUBCUT_FILTER 0
#define CONFIG_ASUPERCUT_FILTER 0
#define CONFIG_ASUPERPASS_FILTER 0
#define CONFIG_ASUPERSTOP_FILTER 0
#define CONFIG_ATEMPO_FILTER 0
#define CONFIG_ATILT_FILTER 0
#define CONFIG_ATRIM_FILTER 0
#define CONFIG_AXCORRELATE_FILTER 0
#define CONFIG_AZMQ_FILTER 0
#define CONFIG_BANDPASS_FILTER 0
#define CONFIG_BANDREJECT_FILTER 0
#define CONFIG_BASS_FILTER 0
#define CONFIG_BIQUAD_FILTER 0
#define CONFIG_BS2B_FILTER 0
#define CONFIG_CHANNELMAP_FILTER 0
#define CONFIG_CHANNELSPLIT_FILTER 0
#define CONFIG_CHORUS_FILTER 0
#define CONFIG_COMPAND_FILTER 0
#define CONFIG_COMPENSATIONDELAY_FILTER 0
#define CONFIG_CROSSFEED_FILTER 0
#define CONFIG_CRYSTALIZER_FILTER 0
#define CONFIG_DCSHIFT_FILTER 0
#define CONFIG_DEESSER_FILTER 0
#define CONFIG_DIALOGUENHANCE_FILTER 0
#define CONFIG_DRMETER_FILTER 0
#define CONFIG_DYNAUDNORM_FILTER 0
#define CONFIG_EARWAX_FILTER 0
#define CONFIG_EBUR128_FILTER 0
#define CONFIG_EQUALIZER_FILTER 0
#define CONFIG_EXTRASTEREO_FILTER 0
#define CONFIG_FIREQUALIZER_FILTER 0
#define CONFIG_FLANGER_FILTER 0
#define CONFIG_HAAS_FILTER 0
#define CONFIG_HDCD_FILTER 0
#define CONFIG_HEADPHONE_FILTER 0
#define CONFIG_HIGHPASS_FILTER 0
#define CONFIG_HIGHSHELF_FILTER 0
#define CONFIG_JOIN_FILTER 0
#define CONFIG_LADSPA_FILTER 0
#define CONFIG_LOUDNORM_FILTER 0
#define CONFIG_LOWPASS_FILTER 0
#define CONFIG_LOWSHELF_FILTER 0
#define CONFIG_LV2_FILTER 0
#define CONFIG_MCOMPAND_FILTER 0
#define CONFIG_PAN_FILTER 0
#define CONFIG_REPLAYGAIN_FILTER 0
#define CONFIG_RUBBERBAND_FILTER 0
#define CONFIG_SIDECHAINCOMPRESS_FILTER 0
#define CONFIG_SIDECHAINGATE_FILTER 0
#define CONFIG_SILENCEDETECT_FILTER 0
#define CONFIG_SILENCEREMOVE_FILTER 0
#define CONFIG_SOFALIZER_FILTER 0
#define CONFIG_SPEECHNORM_FILTER 0
#define CONFIG_STEREOTOOLS_FILTER 0
#define CONFIG_STEREOWIDEN_FILTER 0
#define CONFIG_SUPEREQUALIZER_FILTER 0
#define CONFIG_SURROUND_FILTER 0
#define CONFIG_TILTSHELF_FILTER 0
#define CONFIG_TREBLE_FILTER 0
#define CONFIG_TREMOLO_FILTER 0
#define CONFIG_VIBRATO_FILTER 0
#define CONFIG_VIRTUALBASS_FILTER 0
#define CONFIG_VOLUME_FILTER 0
#define CONFIG_VOLUMEDETECT_FILTER 0
#define CONFIG_WHISPER_FILTER 0
#define CONFIG_AEVALSRC_FILTER 0
#define CONFIG_AFDELAYSRC_FILTER 0
#define CONFIG_AFIREQSRC_FILTER 0
#define CONFIG_AFIRSRC_FILTER 0
#define CONFIG_ANOISESRC_FILTER 0
#define CONFIG_ANULLSRC_FILTER 0
#define CONFIG_FLITE_FILTER 0
#define CONFIG_HILBERT_FILTER 0
#define CONFIG_SINC_FILTER 0
#define CONFIG_SINE_FILTER 0
#define CONFIG_ANULLSINK_FILTER 0
#define CONFIG_ADDROI_FILTER 0
#define CONFIG_ALPHAEXTRACT_FILTER 0
#define CONFIG_ALPHAMERGE_FILTER 0
#define CONFIG_AMPLIFY_FILTER 0
#define CONFIG_ASS_FILTER 0
#define CONFIG_ATADENOISE_FILTER 0
#define CONFIG_AVGBLUR_FILTER 0
#define CONFIG_AVGBLUR_OPENCL_FILTER 0
#define CONFIG_AVGBLUR_VULKAN_FILTER 0
#define CONFIG_BACKGROUNDKEY_FILTER 0
#define CONFIG_BBOX_FILTER 0
#define CONFIG_BENCH_FILTER 0
#define CONFIG_BILATERAL_FILTER 0
#define CONFIG_BILATERAL_CUDA_FILTER 0
#define CONFIG_BITPLANENOISE_FILTER 0
#define CONFIG_BLACKDETECT_FILTER 0
#define CONFIG_BLACKDETECT_VULKAN_FILTER 0
#define CONFIG_BLACKFRAME_FILTER 0
#define CONFIG_BLEND_FILTER 0
#define CONFIG_BLEND_VULKAN_FILTER 0
#define CONFIG_BLOCKDETECT_FILTER 0
#define CONFIG_BLURDETECT_FILTER 0
#define CONFIG_BM3D_FILTER 0
#define CONFIG_BOXBLUR_FILTER 0
#define CONFIG_BOXBLUR_OPENCL_FILTER 0
#define CONFIG_BWDIF_FILTER 0
#define CONFIG_BWDIF_CUDA_FILTER 0
#define CONFIG_BWDIF_VULKAN_FILTER 0
#define CONFIG_CAS_FILTER 0
#define CONFIG_CCREPACK_FILTER 0
#define CONFIG_CHROMABER_VULKAN_FILTER 0
#define CONFIG_CHROMAHOLD_FILTER 0
#define CONFIG_CHROMAKEY_FILTER 0
#define CONFIG_CHROMAKEY_CUDA_FILTER 0
#define CONFIG_CHROMANR_FILTER 0
#define CONFIG_CHROMASHIFT_FILTER 0
#define CONFIG_CIESCOPE_FILTER 0
#define CONFIG_CODECVIEW_FILTER 0
#define CONFIG_COLORBALANCE_FILTER 0
#define CONFIG_COLORCHANNELMIXER_FILTER 0
#define CONFIG_COLORCONTRAST_FILTER 0
#define CONFIG_COLORCORRECT_FILTER 0
#define CONFIG_COLORDETECT_FILTER 0
#define CONFIG_COLORIZE_FILTER 0
#define CONFIG_COLORKEY_FILTER 0
#define CONFIG_COLORKEY_OPENCL_FILTER 0
#define CONFIG_COLORHOLD_FILTER 0
#define CONFIG_COLORLEVELS_FILTER 0
#define CONFIG_COLORMAP_FILTER 0
#define CONFIG_COLORMATRIX_FILTER 0
#define CONFIG_COLORSPACE_FILTER 0
#define CONFIG_COLORSPACE_CUDA_FILTER 0
#define CONFIG_COLORTEMPERATURE_FILTER 0
#define CONFIG_CONVOLUTION_FILTER 0
#define CONFIG_CONVOLUTION_OPENCL_FILTER 0
#define CONFIG_CONVOLVE_FILTER 0
#define CONFIG_COPY_FILTER 0
#define CONFIG_COREIMAGE_FILTER 0
#define CONFIG_CORR_FILTER 0
#define CONFIG_COVER_RECT_FILTER 0
#define CONFIG_CROP_FILTER 0
#define CONFIG_CROPDETECT_FILTER 0
#define CONFIG_CUE_FILTER 0
#define CONFIG_CURVES_FILTER 0
#define CONFIG_DATASCOPE_FILTER 0
#define CONFIG_DBLUR_FILTER 0
#define CONFIG_DCTDNOIZ_FILTER 0
#define CONFIG_DEBAND_FILTER 0
#define CONFIG_DEBLOCK_FILTER 0
#define CONFIG_DECIMATE_FILTER 0
#define CONFIG_DECONVOLVE_FILTER 0
#define CONFIG_DEDOT_FILTER 0
#define CONFIG_DEFLATE_FILTER 0
#define CONFIG_DEFLICKER_FILTER 0
#define CONFIG_DEINTERLACE_QSV_FILTER 0
#define CONFIG_DEINTERLACE_D3D12_FILTER 0
#define CONFIG_DEINTERLACE_VAAPI_FILTER 0
#define CONFIG_DEJUDDER_FILTER 0
#define CONFIG_DELOGO_FILTER 0
#define CONFIG_DENOISE_VAAPI_FILTER 0
#define CONFIG_DERAIN_FILTER 0
#define CONFIG_DESHAKE_FILTER 0
#define CONFIG_DESHAKE_OPENCL_FILTER 0
#define CONFIG_DESPILL_FILTER 0
#define CONFIG_DETELECINE_FILTER 0
#define CONFIG_DILATION_FILTER 0
#define CONFIG_DILATION_OPENCL_FILTER 0
#define CONFIG_DISPLACE_FILTER 0
#define CONFIG_DNN_CLASSIFY_FILTER 0
#define CONFIG_DNN_DETECT_FILTER 0
#define CONFIG_DNN_PROCESSING_FILTER 0
#define CONFIG_DOUBLEWEAVE_FILTER 0
#define CONFIG_DRAWBOX_FILTER 0
#define CONFIG_DRAWGRAPH_FILTER 0
#define CONFIG_DRAWGRID_FILTER 0
#define CONFIG_DRAWTEXT_FILTER 0
#define CONFIG_DRAWVG_FILTER 0
#define CONFIG_EDGEDETECT_FILTER 0
#define CONFIG_ELBG_FILTER 0
#define CONFIG_ENTROPY_FILTER 0
#define CONFIG_EPX_FILTER 0
#define CONFIG_EQ_FILTER 0
#define CONFIG_EROSION_FILTER 0
#define CONFIG_EROSION_OPENCL_FILTER 0
#define CONFIG_ESTDIF_FILTER 0
#define CONFIG_EXPOSURE_FILTER 0
#define CONFIG_EXTRACTPLANES_FILTER 0
#define CONFIG_FADE_FILTER 0
#define CONFIG_FEEDBACK_FILTER 0
#define CONFIG_FFTDNOIZ_FILTER 0
#define CONFIG_FFTFILT_FILTER 0
#define CONFIG_FIELD_FILTER 0
#define CONFIG_FIELDHINT_FILTER 0
#define CONFIG_FIELDMATCH_FILTER 0
#define CONFIG_FIELDORDER_FILTER 0
#define CONFIG_FILLBORDERS_FILTER 0
#define CONFIG_FIND_RECT_FILTER 0
#define CONFIG_FLIP_VULKAN_FILTER 0
#define CONFIG_FLOODFILL_FILTER 0
#define CONFIG_FORMAT_FILTER 0
#define CONFIG_FPS_FILTER 0
#define CONFIG_FRAMEPACK_FILTER 0
#define CONFIG_FRAMERATE_FILTER 0
#define CONFIG_FRAMESTEP_FILTER 0
#define CONFIG_FREEZEDETECT_FILTER 0
#define CONFIG_FREEZEFRAMES_FILTER 0
#define CONFIG_FREI0R_FILTER 0
#define CONFIG_FSPP_FILTER 0
#define CONFIG_FSYNC_FILTER 0
#define CONFIG_GBLUR_FILTER 0
#define CONFIG_GBLUR_VULKAN_FILTER 0
#define CONFIG_GEQ_FILTER 0
#define CONFIG_GRADFUN_FILTER 0
#define CONFIG_GRAPHMONITOR_FILTER 0
#define CONFIG_GRAYWORLD_FILTER 0
#define CONFIG_GREYEDGE_FILTER 0
#define CONFIG_GUIDED_FILTER 0
#define CONFIG_HALDCLUT_FILTER 0
#define CONFIG_HFLIP_FILTER 0
#define CONFIG_HFLIP_VULKAN_FILTER 0
#define CONFIG_HISTEQ_FILTER 0
#define CONFIG_HISTOGRAM_FILTER 0
#define CONFIG_HQDN3D_FILTER 0
#define CONFIG_HQX_FILTER 0
#define CONFIG_HSTACK_FILTER 0
#define CONFIG_HSVHOLD_FILTER 0
#define CONFIG_HSVKEY_FILTER 0
#define CONFIG_HUE_FILTER 0
#define CONFIG_HUESATURATION_FILTER 0
#define CONFIG_HWDOWNLOAD_FILTER 0
#define CONFIG_HWMAP_FILTER 0
#define CONFIG_HWUPLOAD_FILTER 0
#define CONFIG_HWUPLOAD_CUDA_FILTER 0
#define CONFIG_HYSTERESIS_FILTER 0
#define CONFIG_ICCDETECT_FILTER 0
#define CONFIG_ICCGEN_FILTER 0
#define CONFIG_IDENTITY_FILTER 0
#define CONFIG_IDET_FILTER 0
#define CONFIG_IL_FILTER 0
#define CONFIG_INFLATE_FILTER 0
#define CONFIG_INTERLACE_FILTER 0
#define CONFIG_INTERLACE_VULKAN_FILTER 0
#define CONFIG_INTERLEAVE_FILTER 0
#define CONFIG_KERNDEINT_FILTER 0
#define CONFIG_KIRSCH_FILTER 0
#define CONFIG_LAGFUN_FILTER 0
#define CONFIG_LATENCY_FILTER 0
#define CONFIG_LCEVC_FILTER 0
#define CONFIG_LENSCORRECTION_FILTER 0
#define CONFIG_LENSFUN_FILTER 0
#define CONFIG_LIBPLACEBO_FILTER 0
#define CONFIG_LIBVMAF_FILTER 0
#define CONFIG_LIBVMAF_CUDA_FILTER 0
#define CONFIG_LIMITDIFF_FILTER 0
#define CONFIG_LIMITER_FILTER 0
#define CONFIG_LOOP_FILTER 0
#define CONFIG_LUMAKEY_FILTER 0
#define CONFIG_LUT_FILTER 0
#define CONFIG_LUT1D_FILTER 0
#define CONFIG_LUT2_FILTER 0
#define CONFIG_LUT3D_FILTER 0
#define CONFIG_LUTRGB_FILTER 0
#define CONFIG_LUTYUV_FILTER 0
#define CONFIG_MASKEDCLAMP_FILTER 0
#define CONFIG_MASKEDMAX_FILTER 0
#define CONFIG_MASKEDMERGE_FILTER 0
#define CONFIG_MASKEDMIN_FILTER 0
#define CONFIG_MASKEDTHRESHOLD_FILTER 0
#define CONFIG_MASKFUN_FILTER 0
#define CONFIG_MCDEINT_FILTER 0
#define CONFIG_MEDIAN_FILTER 0
#define CONFIG_MERGEPLANES_FILTER 0
#define CONFIG_MESTIMATE_FILTER 0
#define CONFIG_MESTIMATE_D3D12_FILTER 0
#define CONFIG_METADATA_FILTER 0
#define CONFIG_MIDEQUALIZER_FILTER 0
#define CONFIG_MINTERPOLATE_FILTER 0
#define CONFIG_MIX_FILTER 0
#define CONFIG_MONOCHROME_FILTER 0
#define CONFIG_MORPHO_FILTER 0
#define CONFIG_MPDECIMATE_FILTER 0
#define CONFIG_MSAD_FILTER 0
#define CONFIG_MULTIPLY_FILTER 0
#define CONFIG_NEGATE_FILTER 0
#define CONFIG_NLMEANS_FILTER 0
#define CONFIG_NLMEANS_OPENCL_FILTER 0
#define CONFIG_NLMEANS_VULKAN_FILTER 0
#define CONFIG_NNEDI_FILTER 0
#define CONFIG_NOFORMAT_FILTER 0
#define CONFIG_NOISE_FILTER 0
#define CONFIG_NORMALIZE_FILTER 0
#define CONFIG_NULL_FILTER 0
#define CONFIG_OCR_FILTER 0
#define CONFIG_OCV_FILTER 0
#define CONFIG_OSCILLOSCOPE_FILTER 0
#define CONFIG_OCIO_FILTER 0
#define CONFIG_OVERLAY_FILTER 0
#define CONFIG_OVERLAY_OPENCL_FILTER 0
#define CONFIG_OVERLAY_QSV_FILTER 0
#define CONFIG_OVERLAY_VAAPI_FILTER 0
#define CONFIG_OVERLAY_VULKAN_FILTER 0
#define CONFIG_OVERLAY_CUDA_FILTER 0
#define CONFIG_OWDENOISE_FILTER 0
#define CONFIG_PAD_FILTER 0
#define CONFIG_PAD_CUDA_FILTER 0
#define CONFIG_PAD_OPENCL_FILTER 0
#define CONFIG_PALETTEGEN_FILTER 0
#define CONFIG_PALETTEUSE_FILTER 0
#define CONFIG_PERMS_FILTER 0
#define CONFIG_PERSPECTIVE_FILTER 0
#define CONFIG_PHASE_FILTER 0
#define CONFIG_PHOTOSENSITIVITY_FILTER 0
#define CONFIG_PIXDESCTEST_FILTER 0
#define CONFIG_PIXELIZE_FILTER 0
#define CONFIG_PIXSCOPE_FILTER 0
#define CONFIG_PP7_FILTER 0
#define CONFIG_PREMULTIPLY_FILTER 0
#define CONFIG_PREMULTIPLY_DYNAMIC_FILTER 0
#define CONFIG_PREWITT_FILTER 0
#define CONFIG_PREWITT_OPENCL_FILTER 0
#define CONFIG_PROCAMP_VAAPI_FILTER 0
#define CONFIG_PROGRAM_OPENCL_FILTER 0
#define CONFIG_PSEUDOCOLOR_FILTER 0
#define CONFIG_PSNR_FILTER 0
#define CONFIG_PULLUP_FILTER 0
#define CONFIG_QP_FILTER 0
#define CONFIG_QRENCODE_FILTER 0
#define CONFIG_QUIRC_FILTER 0
#define CONFIG_RANDOM_FILTER 0
#define CONFIG_READEIA608_FILTER 0
#define CONFIG_READVITC_FILTER 0
#define CONFIG_REALTIME_FILTER 0
#define CONFIG_REMAP_FILTER 0
#define CONFIG_REMAP_OPENCL_FILTER 0
#define CONFIG_REMOVEGRAIN_FILTER 0
#define CONFIG_REMOVELOGO_FILTER 0
#define CONFIG_REPEATFIELDS_FILTER 0
#define CONFIG_REVERSE_FILTER 0
#define CONFIG_RGBASHIFT_FILTER 0
#define CONFIG_ROBERTS_FILTER 0
#define CONFIG_ROBERTS_OPENCL_FILTER 0
#define CONFIG_ROTATE_FILTER 0
#define CONFIG_SAB_FILTER 0
#define CONFIG_SCALE_FILTER 0
#define CONFIG_VPP_AMF_FILTER 0
#define CONFIG_SR_AMF_FILTER 0
#define CONFIG_SCALE_CUDA_FILTER 0
#define CONFIG_SCALE_D3D11_FILTER 0
#define CONFIG_SCALE_D3D12_FILTER 0
#define CONFIG_SCALE_NPP_FILTER 0
#define CONFIG_SCALE_QSV_FILTER 0
#define CONFIG_SCALE_VAAPI_FILTER 0
#define CONFIG_SCALE_VT_FILTER 0
#define CONFIG_SCALE_VULKAN_FILTER 0
#define CONFIG_SCALE2REF_FILTER 0
#define CONFIG_SCALE2REF_NPP_FILTER 0
#define CONFIG_SCDET_FILTER 0
#define CONFIG_SCDET_VULKAN_FILTER 0
#define CONFIG_SCHARR_FILTER 0
#define CONFIG_SCROLL_FILTER 0
#define CONFIG_SEGMENT_FILTER 0
#define CONFIG_SELECT_FILTER 0
#define CONFIG_SELECTIVECOLOR_FILTER 0
#define CONFIG_SENDCMD_FILTER 0
#define CONFIG_SEPARATEFIELDS_FILTER 0
#define CONFIG_SETDAR_FILTER 0
#define CONFIG_SETFIELD_FILTER 0
#define CONFIG_SETPARAMS_FILTER 0
#define CONFIG_SETPTS_FILTER 0
#define CONFIG_SETRANGE_FILTER 0
#define CONFIG_SETSAR_FILTER 0
#define CONFIG_SETTB_FILTER 0
#define CONFIG_SHARPEN_NPP_FILTER 0
#define CONFIG_SHARPNESS_VAAPI_FILTER 0
#define CONFIG_SHEAR_FILTER 0
#define CONFIG_SHOWINFO_FILTER 0
#define CONFIG_SHOWPALETTE_FILTER 0
#define CONFIG_SHUFFLEFRAMES_FILTER 0
#define CONFIG_SHUFFLEPIXELS_FILTER 0
#define CONFIG_SHUFFLEPLANES_FILTER 0
#define CONFIG_SIDEDATA_FILTER 0
#define CONFIG_SIGNALSTATS_FILTER 0
#define CONFIG_SIGNATURE_FILTER 0
#define CONFIG_SITI_FILTER 0
#define CONFIG_SMARTBLUR_FILTER 0
#define CONFIG_SOBEL_FILTER 0
#define CONFIG_SOBEL_OPENCL_FILTER 0
#define CONFIG_SPLIT_FILTER 0
#define CONFIG_SPP_FILTER 0
#define CONFIG_SR_FILTER 0
#define CONFIG_SSIM_FILTER 0
#define CONFIG_SSIM360_FILTER 0
#define CONFIG_STEREO3D_FILTER 0
#define CONFIG_STREAMSELECT_FILTER 0
#define CONFIG_SUBTITLES_FILTER 0
#define CONFIG_SUPER2XSAI_FILTER 0
#define CONFIG_SWAPRECT_FILTER 0
#define CONFIG_SWAPUV_FILTER 0
#define CONFIG_TBLEND_FILTER 0
#define CONFIG_TELECINE_FILTER 0
#define CONFIG_THISTOGRAM_FILTER 0
#define CONFIG_THRESHOLD_FILTER 0
#define CONFIG_THUMBNAIL_FILTER 0
#define CONFIG_THUMBNAIL_CUDA_FILTER 0
#define CONFIG_TILE_FILTER 0
#define CONFIG_TILTANDSHIFT_FILTER 0
#define CONFIG_TINTERLACE_FILTER 0
#define CONFIG_TLUT2_FILTER 0
#define CONFIG_TMEDIAN_FILTER 0
#define CONFIG_TMIDEQUALIZER_FILTER 0
#define CONFIG_TMIX_FILTER 0
#define CONFIG_TONEMAP_FILTER 0
#define CONFIG_TONEMAP_OPENCL_FILTER 0
#define CONFIG_TONEMAP_VAAPI_FILTER 0
#define CONFIG_TPAD_FILTER 0
#define CONFIG_TRANSPOSE_FILTER 0
#define CONFIG_TRANSPOSE_NPP_FILTER 0
#define CONFIG_TRANSPOSE_OPENCL_FILTER 0
#define CONFIG_TRANSPOSE_VAAPI_FILTER 0
#define CONFIG_TRANSPOSE_VT_FILTER 0
#define CONFIG_TRANSPOSE_VULKAN_FILTER 0
#define CONFIG_TRIM_FILTER 0
#define CONFIG_UNPREMULTIPLY_FILTER 0
#define CONFIG_UNSHARP_FILTER 0
#define CONFIG_UNSHARP_OPENCL_FILTER 0
#define CONFIG_UNTILE_FILTER 0
#define CONFIG_USPP_FILTER 0
#define CONFIG_V360_FILTER 0
#define CONFIG_VAGUEDENOISER_FILTER 0
#define CONFIG_VARBLUR_FILTER 0
#define CONFIG_VECTORSCOPE_FILTER 0
#define CONFIG_VFLIP_FILTER 0
#define CONFIG_VFLIP_VULKAN_FILTER 0
#define CONFIG_VFRDET_FILTER 0
#define CONFIG_VIBRANCE_FILTER 0
#define CONFIG_VIDSTABDETECT_FILTER 0
#define CONFIG_VIDSTABTRANSFORM_FILTER 0
#define CONFIG_VIF_FILTER 0
#define CONFIG_VIGNETTE_FILTER 0
#define CONFIG_VMAFMOTION_FILTER 0
#define CONFIG_VPP_QSV_FILTER 0
#define CONFIG_VSTACK_FILTER 0
#define CONFIG_W3FDIF_FILTER 0
#define CONFIG_WAVEFORM_FILTER 0
#define CONFIG_WEAVE_FILTER 0
#define CONFIG_XBR_FILTER 0
#define CONFIG_XCORRELATE_FILTER 0
#define CONFIG_XFADE_FILTER 0
#define CONFIG_XFADE_OPENCL_FILTER 0
#define CONFIG_XFADE_VULKAN_FILTER 0
#define CONFIG_XMEDIAN_FILTER 0
#define CONFIG_XPSNR_FILTER 0
#define CONFIG_XSTACK_FILTER 0
#define CONFIG_YADIF_FILTER 0
#define CONFIG_YADIF_CUDA_FILTER 0
#define CONFIG_YADIF_VIDEOTOOLBOX_FILTER 0
#define CONFIG_YAEPBLUR_FILTER 0
#define CONFIG_ZMQ_FILTER 0
#define CONFIG_ZOOMPAN_FILTER 0
#define CONFIG_ZSCALE_FILTER 0
#define CONFIG_HSTACK_VAAPI_FILTER 0
#define CONFIG_VSTACK_VAAPI_FILTER 0
#define CONFIG_XSTACK_VAAPI_FILTER 0
#define CONFIG_HSTACK_QSV_FILTER 0
#define CONFIG_VSTACK_QSV_FILTER 0
#define CONFIG_XSTACK_QSV_FILTER 0
#define CONFIG_PAD_VAAPI_FILTER 0
#define CONFIG_DRAWBOX_VAAPI_FILTER 0
#define CONFIG_ALLRGB_FILTER 0
#define CONFIG_ALLYUV_FILTER 0
#define CONFIG_AMF_CAPTURE_FILTER 0
#define CONFIG_CELLAUTO_FILTER 0
#define CONFIG_COLOR_FILTER 0
#define CONFIG_COLOR_VULKAN_FILTER 0
#define CONFIG_COLORCHART_FILTER 0
#define CONFIG_COLORSPECTRUM_FILTER 0
#define CONFIG_COREIMAGESRC_FILTER 0
#define CONFIG_DDAGRAB_FILTER 0
#define CONFIG_FREI0R_SRC_FILTER 0
#define CONFIG_GFXCAPTURE_FILTER 0
#define CONFIG_GRADIENTS_FILTER 0
#define CONFIG_HALDCLUTSRC_FILTER 0
#define CONFIG_LIFE_FILTER 0
#define CONFIG_MANDELBROT_FILTER 0
#define CONFIG_MPTESTSRC_FILTER 0
#define CONFIG_NULLSRC_FILTER 0
#define CONFIG_OPENCLSRC_FILTER 0
#define CONFIG_QRENCODESRC_FILTER 0
#define CONFIG_PAL75BARS_FILTER 0
#define CONFIG_PAL100BARS_FILTER 0
#define CONFIG_PERLIN_FILTER 0
#define CONFIG_RGBTESTSRC_FILTER 0
#define CONFIG_SIERPINSKI_FILTER 0
#define CONFIG_SMPTEBARS_FILTER 0
#define CONFIG_SMPTEHDBARS_FILTER 0
#define CONFIG_TESTSRC_FILTER 0
#define CONFIG_TESTSRC2_FILTER 0
#define CONFIG_YUVTESTSRC_FILTER 0
#define CONFIG_ZONEPLATE_FILTER 0
#define CONFIG_NULLSINK_FILTER 0
#define CONFIG_A3DSCOPE_FILTER 0
#define CONFIG_ABITSCOPE_FILTER 0
#define CONFIG_ADRAWGRAPH_FILTER 0
#define CONFIG_AGRAPHMONITOR_FILTER 0
#define CONFIG_AHISTOGRAM_FILTER 0
#define CONFIG_APHASEMETER_FILTER 0
#define CONFIG_AVECTORSCOPE_FILTER 0
#define CONFIG_CONCAT_FILTER 0
#define CONFIG_SHOWCQT_FILTER 0
#define CONFIG_SHOWCWT_FILTER 0
#define CONFIG_SHOWFREQS_FILTER 0
#define CONFIG_SHOWSPATIAL_FILTER 0
#define CONFIG_SHOWSPECTRUM_FILTER 0
#define CONFIG_SHOWSPECTRUMPIC_FILTER 0
#define CONFIG_SHOWVOLUME_FILTER 0
#define CONFIG_SHOWWAVES_FILTER 0
#define CONFIG_SHOWWAVESPIC_FILTER 0
#define CONFIG_SPECTRUMSYNTH_FILTER 0
#define CONFIG_AVSYNCTEST_FILTER 0
#define CONFIG_AMOVIE_FILTER 0
#define CONFIG_MOVIE_FILTER 0
#define CONFIG_AA_DEMUXER 0
#define CONFIG_AAC_DEMUXER 1
#define CONFIG_AAX_DEMUXER 0
#define CONFIG_AC3_DEMUXER 0
#define CONFIG_AC4_DEMUXER 0
#define CONFIG_ACE_DEMUXER 0
#define CONFIG_ACM_DEMUXER 0
#define CONFIG_ACT_DEMUXER 0
#define CONFIG_ADF_DEMUXER 0
#define CONFIG_ADP_DEMUXER 0
#define CONFIG_ADS_DEMUXER 0
#define CONFIG_ADX_DEMUXER 0
#define CONFIG_AEA_DEMUXER 0
#define CONFIG_AFC_DEMUXER 0
#define CONFIG_AIFF_DEMUXER 0
#define CONFIG_AIX_DEMUXER 0
#define CONFIG_ALP_DEMUXER 0
#define CONFIG_AMR_DEMUXER 0
#define CONFIG_AMRNB_DEMUXER 0
#define CONFIG_AMRWB_DEMUXER 0
#define CONFIG_ANM_DEMUXER 0
#define CONFIG_APAC_DEMUXER 0
#define CONFIG_APC_DEMUXER 0
#define CONFIG_APE_DEMUXER 0
#define CONFIG_APM_DEMUXER 0
#define CONFIG_APNG_DEMUXER 0
#define CONFIG_APTX_DEMUXER 0
#define CONFIG_APTX_HD_DEMUXER 0
#define CONFIG_APV_DEMUXER 0
#define CONFIG_AQTITLE_DEMUXER 0
#define CONFIG_ARGO_ASF_DEMUXER 0
#define CONFIG_ARGO_BRP_DEMUXER 0
#define CONFIG_ARGO_CVG_DEMUXER 0
#define CONFIG_ASF_DEMUXER 0
#define CONFIG_ASF_O_DEMUXER 0
#define CONFIG_ASS_DEMUXER 0
#define CONFIG_AST_DEMUXER 0
#define CONFIG_AU_DEMUXER 0
#define CONFIG_AV1_DEMUXER 0
#define CONFIG_AVI_DEMUXER 0
#define CONFIG_AVR_DEMUXER 0
#define CONFIG_AVS_DEMUXER 0
#define CONFIG_AVS2_DEMUXER 0
#define CONFIG_AVS3_DEMUXER 0
#define CONFIG_BETHSOFTVID_DEMUXER 0
#define CONFIG_BFI_DEMUXER 0
#define CONFIG_BINTEXT_DEMUXER 0
#define CONFIG_BINK_DEMUXER 0
#define CONFIG_BINKA_DEMUXER 0
#define CONFIG_BIT_DEMUXER 0
#define CONFIG_BITPACKED_DEMUXER 0
#define CONFIG_BMV_DEMUXER 0
#define CONFIG_BFSTM_DEMUXER 0
#define CONFIG_BRSTM_DEMUXER 0
#define CONFIG_BOA_DEMUXER 0
#define CONFIG_BONK_DEMUXER 0
#define CONFIG_C93_DEMUXER 0
#define CONFIG_CAF_DEMUXER 0
#define CONFIG_CAVSVIDEO_DEMUXER 0
#define CONFIG_CDG_DEMUXER 0
#define CONFIG_CDXL_DEMUXER 0
#define CONFIG_CINE_DEMUXER 0
#define CONFIG_CODEC2_DEMUXER 0
#define CONFIG_CODEC2RAW_DEMUXER 0
#define CONFIG_CONCAT_DEMUXER 0
#define CONFIG_DASH_DEMUXER 0
#define CONFIG_DATA_DEMUXER 0
#define CONFIG_DAUD_DEMUXER 0
#define CONFIG_DCSTR_DEMUXER 0
#define CONFIG_DERF_DEMUXER 0
#define CONFIG_DFA_DEMUXER 0
#define CONFIG_DFPWM_DEMUXER 0
#define CONFIG_DHAV_DEMUXER 0
#define CONFIG_DIRAC_DEMUXER 0
#define CONFIG_DNXHD_DEMUXER 0
#define CONFIG_DSF_DEMUXER 0
#define CONFIG_DSICIN_DEMUXER 0
#define CONFIG_DSS_DEMUXER 0
#define CONFIG_DTS_DEMUXER 0
#define CONFIG_DTSHD_DEMUXER 0
#define CONFIG_DV_DEMUXER 0
#define CONFIG_DVBSUB_DEMUXER 0
#define CONFIG_DVBTXT_DEMUXER 0
#define CONFIG_DXA_DEMUXER 0
#define CONFIG_EA_DEMUXER 0
#define CONFIG_EA_CDATA_DEMUXER 0
#define CONFIG_EAC3_DEMUXER 0
#define CONFIG_EPAF_DEMUXER 0
#define CONFIG_EVC_DEMUXER 0
#define CONFIG_FFMETADATA_DEMUXER 0
#define CONFIG_FILMSTRIP_DEMUXER 0
#define CONFIG_FITS_DEMUXER 0
#define CONFIG_FLAC_DEMUXER 1
#define CONFIG_FLIC_DEMUXER 0
#define CONFIG_FLV_DEMUXER 0
#define CONFIG_LIVE_FLV_DEMUXER 0
#define CONFIG_FOURXM_DEMUXER 0
#define CONFIG_FRM_DEMUXER 0
#define CONFIG_FSB_DEMUXER 0
#define CONFIG_FWSE_DEMUXER 0
#define CONFIG_G722_DEMUXER 0
#define CONFIG_G723_1_DEMUXER 0
#define CONFIG_G726_DEMUXER 0
#define CONFIG_G726LE_DEMUXER 0
#define CONFIG_G728_DEMUXER 0
#define CONFIG_G729_DEMUXER 0
#define CONFIG_GDV_DEMUXER 0
#define CONFIG_GENH_DEMUXER 0
#define CONFIG_GIF_DEMUXER 0
#define CONFIG_GSM_DEMUXER 0
#define CONFIG_GXF_DEMUXER 0
#define CONFIG_H261_DEMUXER 0
#define CONFIG_H263_DEMUXER 0
#define CONFIG_H264_DEMUXER 0
#define CONFIG_HCA_DEMUXER 0
#define CONFIG_HCOM_DEMUXER 0
#define CONFIG_HEVC_DEMUXER 0
#define CONFIG_HLS_DEMUXER 0
#define CONFIG_HNM_DEMUXER 0
#define CONFIG_HXVS_DEMUXER 0
#define CONFIG_IAMF_DEMUXER 0
#define CONFIG_ICO_DEMUXER 0
#define CONFIG_IDCIN_DEMUXER 0
#define CONFIG_IDF_DEMUXER 0
#define CONFIG_IFF_DEMUXER 0
#define CONFIG_IFV_DEMUXER 0
#define CONFIG_ILBC_DEMUXER 0
#define CONFIG_IMAGE2_DEMUXER 0
#define CONFIG_IMAGE2PIPE_DEMUXER 0
#define CONFIG_IMAGE2_ALIAS_PIX_DEMUXER 0
#define CONFIG_IMAGE2_BRENDER_PIX_DEMUXER 0
#define CONFIG_IMF_DEMUXER 0
#define CONFIG_INGENIENT_DEMUXER 0
#define CONFIG_IPMOVIE_DEMUXER 0
#define CONFIG_IPU_DEMUXER 0
#define CONFIG_IRCAM_DEMUXER 0
#define CONFIG_ISS_DEMUXER 0
#define CONFIG_IV8_DEMUXER 0
#define CONFIG_IVF_DEMUXER 0
#define CONFIG_IVR_DEMUXER 0
#define CONFIG_JACOSUB_DEMUXER 0
#define CONFIG_JV_DEMUXER 0
#define CONFIG_JPEGXL_ANIM_DEMUXER 0
#define CONFIG_KUX_DEMUXER 0
#define CONFIG_KVAG_DEMUXER 0
#define CONFIG_LAF_DEMUXER 0
#define CONFIG_LC3_DEMUXER 0
#define CONFIG_LMLM4_DEMUXER 0
#define CONFIG_LOAS_DEMUXER 0
#define CONFIG_LUODAT_DEMUXER 0
#define CONFIG_LRC_DEMUXER 0
#define CONFIG_LVF_DEMUXER 0
#define CONFIG_LXF_DEMUXER 0
#define CONFIG_M4V_DEMUXER 0
#define CONFIG_MCA_DEMUXER 0
#define CONFIG_MCC_DEMUXER 0
#define CONFIG_MATROSKA_DEMUXER 1
#define CONFIG_MGSTS_DEMUXER 0
#define CONFIG_MICRODVD_DEMUXER 0
#define CONFIG_MJPEG_DEMUXER 0
#define CONFIG_MJPEG_2000_DEMUXER 0
#define CONFIG_MLP_DEMUXER 0
#define CONFIG_MLV_DEMUXER 0
#define CONFIG_MM_DEMUXER 0
#define CONFIG_MMF_DEMUXER 0
#define CONFIG_MODS_DEMUXER 0
#define CONFIG_MOFLEX_DEMUXER 0
#define CONFIG_MOV_DEMUXER 1
#define CONFIG_MP3_DEMUXER 1
#define CONFIG_MPC_DEMUXER 0
#define CONFIG_MPC8_DEMUXER 0
#define CONFIG_MPEGPS_DEMUXER 0
#define CONFIG_MPEGTS_DEMUXER 0
#define CONFIG_MPEGTSRAW_DEMUXER 0
#define CONFIG_MPEGVIDEO_DEMUXER 0
#define CONFIG_MPJPEG_DEMUXER 0
#define CONFIG_MPL2_DEMUXER 0
#define CONFIG_MPSUB_DEMUXER 0
#define CONFIG_MSF_DEMUXER 0
#define CONFIG_MSNWC_TCP_DEMUXER 0
#define CONFIG_MSP_DEMUXER 0
#define CONFIG_MTAF_DEMUXER 0
#define CONFIG_MTV_DEMUXER 0
#define CONFIG_MUSX_DEMUXER 0
#define CONFIG_MV_DEMUXER 0
#define CONFIG_MVI_DEMUXER 0
#define CONFIG_MXF_DEMUXER 0
#define CONFIG_MXG_DEMUXER 0
#define CONFIG_NC_DEMUXER 0
#define CONFIG_NISTSPHERE_DEMUXER 0
#define CONFIG_NSP_DEMUXER 0
#define CONFIG_NSV_DEMUXER 0
#define CONFIG_NUT_DEMUXER 0
#define CONFIG_NUV_DEMUXER 0
#define CONFIG_OBU_DEMUXER 0
#define CONFIG_OGG_DEMUXER 1
#define CONFIG_OMA_DEMUXER 0
#define CONFIG_OSQ_DEMUXER 0
#define CONFIG_PAF_DEMUXER 0
#define CONFIG_PCM_ALAW_DEMUXER 0
#define CONFIG_PCM_MULAW_DEMUXER 0
#define CONFIG_PCM_VIDC_DEMUXER 0
#define CONFIG_PCM_F64BE_DEMUXER 0
#define CONFIG_PCM_F64LE_DEMUXER 0
#define CONFIG_PCM_F32BE_DEMUXER 0
#define CONFIG_PCM_F32LE_DEMUXER 0
#define CONFIG_PCM_S32BE_DEMUXER 0
#define CONFIG_PCM_S32LE_DEMUXER 0
#define CONFIG_PCM_S24BE_DEMUXER 0
#define CONFIG_PCM_S24LE_DEMUXER 0
#define CONFIG_PCM_S16BE_DEMUXER 0
#define CONFIG_PCM_S16LE_DEMUXER 0
#define CONFIG_PCM_S8_DEMUXER 0
#define CONFIG_PCM_U32BE_DEMUXER 0
#define CONFIG_PCM_U32LE_DEMUXER 0
#define CONFIG_PCM_U24BE_DEMUXER 0
#define CONFIG_PCM_U24LE_DEMUXER 0
#define CONFIG_PCM_U16BE_DEMUXER 0
#define CONFIG_PCM_U16LE_DEMUXER 0
#define CONFIG_PCM_U8_DEMUXER 0
#define CONFIG_PDV_DEMUXER 0
#define CONFIG_PJS_DEMUXER 0
#define CONFIG_PMP_DEMUXER 0
#define CONFIG_PP_BNK_DEMUXER 0
#define CONFIG_PVA_DEMUXER 0
#define CONFIG_PVF_DEMUXER 0
#define CONFIG_QCP_DEMUXER 0
#define CONFIG_QOA_DEMUXER 0
#define CONFIG_R3D_DEMUXER 0
#define CONFIG_RAWVIDEO_DEMUXER 0
#define CONFIG_RCWT_DEMUXER 0
#define CONFIG_REALTEXT_DEMUXER 0
#define CONFIG_REDSPARK_DEMUXER 0
#define CONFIG_RKA_DEMUXER 0
#define CONFIG_RL2_DEMUXER 0
#define CONFIG_RM_DEMUXER 0
#define CONFIG_ROQ_DEMUXER 0
#define CONFIG_RPL_DEMUXER 0
#define CONFIG_RSD_DEMUXER 0
#define CONFIG_RSO_DEMUXER 0
#define CONFIG_RTP_DEMUXER 0
#define CONFIG_RTSP_DEMUXER 0
#define CONFIG_S337M_DEMUXER 0
#define CONFIG_SAMI_DEMUXER 0
#define CONFIG_SAP_DEMUXER 0
#define CONFIG_SBC_DEMUXER 0
#define CONFIG_SBG_DEMUXER 0
#define CONFIG_SCC_DEMUXER 0
#define CONFIG_SCD_DEMUXER 0
#define CONFIG_SDNS_DEMUXER 0
#define CONFIG_SDP_DEMUXER 0
#define CONFIG_SDR2_DEMUXER 0
#define CONFIG_SDS_DEMUXER 0
#define CONFIG_SDX_DEMUXER 0
#define CONFIG_SEGAFILM_DEMUXER 0
#define CONFIG_SER_DEMUXER 0
#define CONFIG_SGA_DEMUXER 0
#define CONFIG_SHORTEN_DEMUXER 0
#define CONFIG_SIFF_DEMUXER 0
#define CONFIG_SIMBIOSIS_IMX_DEMUXER 0
#define CONFIG_SLN_DEMUXER 0
#define CONFIG_SMACKER_DEMUXER 0
#define CONFIG_SMJPEG_DEMUXER 0
#define CONFIG_SMUSH_DEMUXER 0
#define CONFIG_SOL_DEMUXER 0
#define CONFIG_SOX_DEMUXER 0
#define CONFIG_SPDIF_DEMUXER 0
#define CONFIG_SRT_DEMUXER 0
#define CONFIG_STR_DEMUXER 0
#define CONFIG_STL_DEMUXER 0
#define CONFIG_SUBVIEWER1_DEMUXER 0
#define CONFIG_SUBVIEWER_DEMUXER 0
#define CONFIG_SUP_DEMUXER 0
#define CONFIG_SVAG_DEMUXER 0
#define CONFIG_SVS_DEMUXER 0
#define CONFIG_SWF_DEMUXER 0
#define CONFIG_TAK_DEMUXER 0
#define CONFIG_TEDCAPTIONS_DEMUXER 0
#define CONFIG_THP_DEMUXER 0
#define CONFIG_THREEDOSTR_DEMUXER 0
#define CONFIG_TIERTEXSEQ_DEMUXER 0
#define CONFIG_TMV_DEMUXER 0
#define CONFIG_TRUEHD_DEMUXER 0
#define CONFIG_TTA_DEMUXER 0
#define CONFIG_TXD_DEMUXER 0
#define CONFIG_TTY_DEMUXER 0
#define CONFIG_TY_DEMUXER 0
#define CONFIG_USM_DEMUXER 0
#define CONFIG_V210_DEMUXER 0
#define CONFIG_V210X_DEMUXER 0
#define CONFIG_VAG_DEMUXER 0
#define CONFIG_VC1_DEMUXER 0
#define CONFIG_VC1T_DEMUXER 0
#define CONFIG_VIVIDAS_DEMUXER 0
#define CONFIG_VIVO_DEMUXER 0
#define CONFIG_VMD_DEMUXER 0
#define CONFIG_VOBSUB_DEMUXER 0
#define CONFIG_VOC_DEMUXER 0
#define CONFIG_VPK_DEMUXER 0
#define CONFIG_VPLAYER_DEMUXER 0
#define CONFIG_VQF_DEMUXER 0
#define CONFIG_VVC_DEMUXER 0
#define CONFIG_W64_DEMUXER 0
#define CONFIG_WADY_DEMUXER 0
#define CONFIG_WAVARC_DEMUXER 0
#define CONFIG_WAV_DEMUXER 1
#define CONFIG_WC3_DEMUXER 0
#define CONFIG_WEBM_DASH_MANIFEST_DEMUXER 0
#define CONFIG_WEBVTT_DEMUXER 0
#define CONFIG_WSAUD_DEMUXER 0
#define CONFIG_WSD_DEMUXER 0
#define CONFIG_WSVQA_DEMUXER 0
#define CONFIG_WTV_DEMUXER 0
#define CONFIG_WVE_DEMUXER 0
#define CONFIG_WV_DEMUXER 0
#define CONFIG_XA_DEMUXER 0
#define CONFIG_XBIN_DEMUXER 0
#define CONFIG_XMD_DEMUXER 0
#define CONFIG_XMV_DEMUXER 0
#define CONFIG_XVAG_DEMUXER 0
#define CONFIG_XWMA_DEMUXER 0
#define CONFIG_YOP_DEMUXER 0
#define CONFIG_YUV4MPEGPIPE_DEMUXER 0
#define CONFIG_IMAGE_BMP_PIPE_DEMUXER 0
#define CONFIG_IMAGE_CRI_PIPE_DEMUXER 0
#define CONFIG_IMAGE_DDS_PIPE_DEMUXER 0
#define CONFIG_IMAGE_DPX_PIPE_DEMUXER 0
#define CONFIG_IMAGE_EXR_PIPE_DEMUXER 0
#define CONFIG_IMAGE_GEM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_GIF_PIPE_DEMUXER 0
#define CONFIG_IMAGE_HDR_PIPE_DEMUXER 0
#define CONFIG_IMAGE_J2K_PIPE_DEMUXER 0
#define CONFIG_IMAGE_JPEG_PIPE_DEMUXER 0
#define CONFIG_IMAGE_JPEGLS_PIPE_DEMUXER 0
#define CONFIG_IMAGE_JPEGXL_PIPE_DEMUXER 0
#define CONFIG_IMAGE_JPEGXS_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PAM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PBM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PCX_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PFM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PGMYUV_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PGM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PGX_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PHM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PHOTOCD_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PICTOR_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PNG_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PPM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_PSD_PIPE_DEMUXER 0
#define CONFIG_IMAGE_QDRAW_PIPE_DEMUXER 0
#define CONFIG_IMAGE_QOI_PIPE_DEMUXER 0
#define CONFIG_IMAGE_SGI_PIPE_DEMUXER 0
#define CONFIG_IMAGE_SVG_PIPE_DEMUXER 0
#define CONFIG_IMAGE_SUNRAST_PIPE_DEMUXER 0
#define CONFIG_IMAGE_TIFF_PIPE_DEMUXER 0
#define CONFIG_IMAGE_VBN_PIPE_DEMUXER 0
#define CONFIG_IMAGE_WEBP_PIPE_DEMUXER 0
#define CONFIG_IMAGE_XBM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_XPM_PIPE_DEMUXER 0
#define CONFIG_IMAGE_XWD_PIPE_DEMUXER 0
#define CONFIG_AVISYNTH_DEMUXER 0
#define CONFIG_DVDVIDEO_DEMUXER 0
#define CONFIG_LIBGME_DEMUXER 0
#define CONFIG_LIBMODPLUG_DEMUXER 0
#define CONFIG_LIBOPENMPT_DEMUXER 0
#define CONFIG_VAPOURSYNTH_DEMUXER 0
#define CONFIG_A64_MUXER 0
#define CONFIG_AC3_MUXER 0
#define CONFIG_AC4_MUXER 0
#define CONFIG_ADTS_MUXER 0
#define CONFIG_ADX_MUXER 0
#define CONFIG_AEA_MUXER 0
#define CONFIG_AIFF_MUXER 0
#define CONFIG_ALP_MUXER 0
#define CONFIG_AMR_MUXER 0
#define CONFIG_AMV_MUXER 0
#define CONFIG_APM_MUXER 0
#define CONFIG_APNG_MUXER 0
#define CONFIG_APTX_MUXER 0
#define CONFIG_APTX_HD_MUXER 0
#define CONFIG_APV_MUXER 0
#define CONFIG_ARGO_ASF_MUXER 0
#define CONFIG_ARGO_CVG_MUXER 0
#define CONFIG_ASF_MUXER 0
#define CONFIG_ASS_MUXER 0
#define CONFIG_AST_MUXER 0
#define CONFIG_ASF_STREAM_MUXER 0
#define CONFIG_AU_MUXER 0
#define CONFIG_AVI_MUXER 0
#define CONFIG_AVIF_MUXER 0
#define CONFIG_AVM2_MUXER 0
#define CONFIG_AVS2_MUXER 0
#define CONFIG_AVS3_MUXER 0
#define CONFIG_BIT_MUXER 0
#define CONFIG_CAF_MUXER 0
#define CONFIG_CAVSVIDEO_MUXER 0
#define CONFIG_CODEC2_MUXER 0
#define CONFIG_CODEC2RAW_MUXER 0
#define CONFIG_CRC_MUXER 0
#define CONFIG_DASH_MUXER 0
#define CONFIG_DATA_MUXER 0
#define CONFIG_DAUD_MUXER 0
#define CONFIG_DFPWM_MUXER 0
#define CONFIG_DIRAC_MUXER 0
#define CONFIG_DNXHD_MUXER 0
#define CONFIG_DTS_MUXER 0
#define CONFIG_DV_MUXER 0
#define CONFIG_EAC3_MUXER 0
#define CONFIG_EVC_MUXER 0
#define CONFIG_F4V_MUXER 0
#define CONFIG_FFMETADATA_MUXER 0
#define CONFIG_FIFO_MUXER 0
#define CONFIG_FILMSTRIP_MUXER 0
#define CONFIG_FITS_MUXER 0
#define CONFIG_FLAC_MUXER 0
#define CONFIG_FLV_MUXER 0
#define CONFIG_FRAMECRC_MUXER 0
#define CONFIG_FRAMEHASH_MUXER 0
#define CONFIG_FRAMEMD5_MUXER 0
#define CONFIG_G722_MUXER 0
#define CONFIG_G723_1_MUXER 0
#define CONFIG_G726_MUXER 0
#define CONFIG_G726LE_MUXER 0
#define CONFIG_GIF_MUXER 0
#define CONFIG_GSM_MUXER 0
#define CONFIG_GXF_MUXER 0
#define CONFIG_H261_MUXER 0
#define CONFIG_H263_MUXER 0
#define CONFIG_H264_MUXER 0
#define CONFIG_HASH_MUXER 0
#define CONFIG_HDS_MUXER 0
#define CONFIG_HEVC_MUXER 0
#define CONFIG_HLS_MUXER 0
#define CONFIG_IAMF_MUXER 0
#define CONFIG_ICO_MUXER 0
#define CONFIG_ILBC_MUXER 0
#define CONFIG_IMAGE2_MUXER 0
#define CONFIG_IMAGE2PIPE_MUXER 0
#define CONFIG_IPOD_MUXER 0
#define CONFIG_IRCAM_MUXER 0
#define CONFIG_ISMV_MUXER 0
#define CONFIG_IVF_MUXER 0
#define CONFIG_JACOSUB_MUXER 0
#define CONFIG_KVAG_MUXER 0
#define CONFIG_LATM_MUXER 0
#define CONFIG_LC3_MUXER 0
#define CONFIG_LRC_MUXER 0
#define CONFIG_M4V_MUXER 0
#define CONFIG_MCC_MUXER 0
#define CONFIG_MD5_MUXER 0
#define CONFIG_MATROSKA_MUXER 0
#define CONFIG_MATROSKA_AUDIO_MUXER 0
#define CONFIG_MICRODVD_MUXER 0
#define CONFIG_MJPEG_MUXER 0
#define CONFIG_MLP_MUXER 0
#define CONFIG_MMF_MUXER 0
#define CONFIG_MOV_MUXER 0
#define CONFIG_MP2_MUXER 0
#define CONFIG_MP3_MUXER 0
#define CONFIG_MP4_MUXER 0
#define CONFIG_MPEG1SYSTEM_MUXER 0
#define CONFIG_MPEG1VCD_MUXER 0
#define CONFIG_MPEG1VIDEO_MUXER 0
#define CONFIG_MPEG2DVD_MUXER 0
#define CONFIG_MPEG2SVCD_MUXER 0
#define CONFIG_MPEG2VIDEO_MUXER 0
#define CONFIG_MPEG2VOB_MUXER 0
#define CONFIG_MPEGTS_MUXER 0
#define CONFIG_MPJPEG_MUXER 0
#define CONFIG_MXF_MUXER 0
#define CONFIG_MXF_D10_MUXER 0
#define CONFIG_MXF_OPATOM_MUXER 0
#define CONFIG_NULL_MUXER 0
#define CONFIG_NUT_MUXER 0
#define CONFIG_OBU_MUXER 0
#define CONFIG_OGA_MUXER 0
#define CONFIG_OGG_MUXER 0
#define CONFIG_OGV_MUXER 0
#define CONFIG_OMA_MUXER 0
#define CONFIG_OPUS_MUXER 0
#define CONFIG_PCM_ALAW_MUXER 0
#define CONFIG_PCM_MULAW_MUXER 0
#define CONFIG_PCM_VIDC_MUXER 0
#define CONFIG_PCM_F64BE_MUXER 0
#define CONFIG_PCM_F64LE_MUXER 0
#define CONFIG_PCM_F32BE_MUXER 0
#define CONFIG_PCM_F32LE_MUXER 0
#define CONFIG_PCM_S32BE_MUXER 0
#define CONFIG_PCM_S32LE_MUXER 0
#define CONFIG_PCM_S24BE_MUXER 0
#define CONFIG_PCM_S24LE_MUXER 0
#define CONFIG_PCM_S16BE_MUXER 0
#define CONFIG_PCM_S16LE_MUXER 0
#define CONFIG_PCM_S8_MUXER 0
#define CONFIG_PCM_U32BE_MUXER 0
#define CONFIG_PCM_U32LE_MUXER 0
#define CONFIG_PCM_U24BE_MUXER 0
#define CONFIG_PCM_U24LE_MUXER 0
#define CONFIG_PCM_U16BE_MUXER 0
#define CONFIG_PCM_U16LE_MUXER 0
#define CONFIG_PCM_U8_MUXER 0
#define CONFIG_PSP_MUXER 0
#define CONFIG_RAWVIDEO_MUXER 0
#define CONFIG_RCWT_MUXER 0
#define CONFIG_RM_MUXER 0
#define CONFIG_ROQ_MUXER 0
#define CONFIG_RSO_MUXER 0
#define CONFIG_RTP_MUXER 0
#define CONFIG_RTP_MPEGTS_MUXER 0
#define CONFIG_RTSP_MUXER 0
#define CONFIG_SAP_MUXER 0
#define CONFIG_SBC_MUXER 0
#define CONFIG_SCC_MUXER 0
#define CONFIG_SEGAFILM_MUXER 0
#define CONFIG_SEGMENT_MUXER 0
#define CONFIG_STREAM_SEGMENT_MUXER 0
#define CONFIG_SMJPEG_MUXER 0
#define CONFIG_SMOOTHSTREAMING_MUXER 0
#define CONFIG_SOX_MUXER 0
#define CONFIG_SPX_MUXER 0
#define CONFIG_SPDIF_MUXER 0
#define CONFIG_SRT_MUXER 0
#define CONFIG_STREAMHASH_MUXER 0
#define CONFIG_SUP_MUXER 0
#define CONFIG_SWF_MUXER 0
#define CONFIG_TEE_MUXER 0
#define CONFIG_TG2_MUXER 0
#define CONFIG_TGP_MUXER 0
#define CONFIG_MKVTIMESTAMP_V2_MUXER 0
#define CONFIG_TRUEHD_MUXER 0
#define CONFIG_TTA_MUXER 0
#define CONFIG_TTML_MUXER 0
#define CONFIG_UNCODEDFRAMECRC_MUXER 0
#define CONFIG_VC1_MUXER 0
#define CONFIG_VC1T_MUXER 0
#define CONFIG_VOC_MUXER 0
#define CONFIG_VVC_MUXER 0
#define CONFIG_W64_MUXER 0
#define CONFIG_WAV_MUXER 0
#define CONFIG_WEBM_MUXER 0
#define CONFIG_WEBM_DASH_MANIFEST_MUXER 0
#define CONFIG_WEBM_CHUNK_MUXER 0
#define CONFIG_WEBP_MUXER 0
#define CONFIG_WEBVTT_MUXER 0
#define CONFIG_WHIP_MUXER 0
#define CONFIG_WSAUD_MUXER 0
#define CONFIG_WTV_MUXER 0
#define CONFIG_WV_MUXER 0
#define CONFIG_YUV4MPEGPIPE_MUXER 0
#define CONFIG_CHROMAPRINT_MUXER 0
#define CONFIG_ANDROID_CONTENT_PROTOCOL 0
#define CONFIG_ASYNC_PROTOCOL 0
#define CONFIG_BLURAY_PROTOCOL 0
#define CONFIG_CACHE_PROTOCOL 0
#define CONFIG_CONCAT_PROTOCOL 0
#define CONFIG_CONCATF_PROTOCOL 0
#define CONFIG_CRYPTO_PROTOCOL 0
#define CONFIG_DATA_PROTOCOL 0
#define CONFIG_FD_PROTOCOL 0
#define CONFIG_FFRTMPCRYPT_PROTOCOL 0
#define CONFIG_FFRTMPHTTP_PROTOCOL 0
#define CONFIG_FILE_PROTOCOL 0
#define CONFIG_FTP_PROTOCOL 0
#define CONFIG_GOPHER_PROTOCOL 0
#define CONFIG_GOPHERS_PROTOCOL 0
#define CONFIG_HTTP_PROTOCOL 0
#define CONFIG_HTTPPROXY_PROTOCOL 0
#define CONFIG_HTTPS_PROTOCOL 0
#define CONFIG_ICECAST_PROTOCOL 0
#define CONFIG_MMSH_PROTOCOL 0
#define CONFIG_MMST_PROTOCOL 0
#define CONFIG_MD5_PROTOCOL 0
#define CONFIG_PIPE_PROTOCOL 0
#define CONFIG_PROMPEG_PROTOCOL 0
#define CONFIG_RTMP_PROTOCOL 0
#define CONFIG_RTMPE_PROTOCOL 0
#define CONFIG_RTMPS_PROTOCOL 0
#define CONFIG_RTMPT_PROTOCOL 0
#define CONFIG_RTMPTE_PROTOCOL 0
#define CONFIG_RTMPTS_PROTOCOL 0
#define CONFIG_RTP_PROTOCOL 0
#define CONFIG_SCTP_PROTOCOL 0
#define CONFIG_SRTP_PROTOCOL 0
#define CONFIG_SUBFILE_PROTOCOL 0
#define CONFIG_TEE_PROTOCOL 0
#define CONFIG_TCP_PROTOCOL 0
#define CONFIG_TLS_PROTOCOL 0
#define CONFIG_DTLS_PROTOCOL 0
#define CONFIG_UDP_PROTOCOL 0
#define CONFIG_UDPLITE_PROTOCOL 0
#define CONFIG_UNIX_PROTOCOL 0
#define CONFIG_LIBAMQP_PROTOCOL 0
#define CONFIG_LIBRIST_PROTOCOL 0
#define CONFIG_LIBRTMP_PROTOCOL 0
#define CONFIG_LIBRTMPE_PROTOCOL 0
#define CONFIG_LIBRTMPS_PROTOCOL 0
#define CONFIG_LIBRTMPT_PROTOCOL 0
#define CONFIG_LIBRTMPTE_PROTOCOL 0
#define CONFIG_LIBSRT_PROTOCOL 0
#define CONFIG_LIBSSH_PROTOCOL 0
#define CONFIG_LIBSMBCLIENT_PROTOCOL 0
#define CONFIG_LIBZMQ_PROTOCOL 0
#define CONFIG_IPFS_GATEWAY_PROTOCOL 0
#define CONFIG_IPNS_GATEWAY_PROTOCOL 0
#endif /* FFMPEG_CONFIG_COMPONENTS_H */
